"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""


class SyndicateBaseError(Exception):
    """
    The base exception class for syndicate exceptions.
    """
    pass


# =============================================================================
class NotImplementedError(SyndicateBaseError):
    """
    The syndicate not implemented error.
    """
    pass


class InternalError(SyndicateBaseError):
    """
    The syndicate internal error.
    """
    pass


class AbortedError(SyndicateBaseError):
    """
    The operation abortion error.
    """
    pass


class InvalidValueError(SyndicateBaseError):
    """
    The value error.
    """
    pass


class InvalidTypeError(SyndicateBaseError):
    """
    The type error.
    """
    pass


# =============================================================================
class ProjectStateError(SyndicateBaseError):
    """
    Project state error.
    """
    pass


# =============================================================================
class ConfigurationError(SyndicateBaseError):
    """
    The syndicate configuration error.
    """
    pass


# =============================================================================
class EnvironmentError(SyndicateBaseError):
    """
    The syndicate environment error.
    """
    pass


# =============================================================================
class ResourceProcessingError(SyndicateBaseError):
    """
    The resource processing error.
    """
    pass


class ArtifactError(ResourceProcessingError):
    """
    The artifact error.
    """
    pass


class ArtifactAssemblingError(ResourceProcessingError):
    """
    The artifact error.
    """
    pass


class ResourceNotFoundError(ResourceProcessingError):
    """
    The resource cannot be found.
    """
    pass


# -----------------------------------------------------------------------------
class ResourceMetadataError(ResourceProcessingError):
    """
    The resource metadata error.
    """
    pass


class ParameterError(ResourceMetadataError):
    """
    The resource metadata parameter error.
    """
    pass


# =============================================================================



from syndicate.core.constants import BATCH_JOBDEF_TYPE
from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator


class BatchJobdefGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = BATCH_JOBDEF_TYPE
    CONFIGURATION = {
        'job_definition_type': None,
        'container_properties': {
            'image': 'alpine',
            'vcpus': 1,
            'memory': 1024,
            'readonly_root_filesystem': True,
            'command': list,
            'job_role_arn': None
        },
        'node_properties': dict,
        'retry_strategy': dict,
        'tags': dict
    }



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import click

from syndicate.commons.log_helper import get_user_logger
from syndicate.core.constants import OK_RETURN_CODE, FAILED_RETURN_CODE
from syndicate.core.decorators import return_code_manager, tags_to_context
from syndicate.core.helper import verbose_option

TAGS_GROUP_NAME = 'tags'


USER_LOG = get_user_logger()


@click.group(name=TAGS_GROUP_NAME)
@return_code_manager
def tags():
    """Manage resources tags"""


@tags.command(name='apply')
@return_code_manager
@tags_to_context
@verbose_option
@click.pass_context
def apply(ctx):
    """Assign tags from config to deployed resources"""
    output = ctx.obj['output']
    success = ctx.obj['tags_resource'].safe_apply_tags(output)
    if success:
        USER_LOG.info("Tags applied successfully")
        return OK_RETURN_CODE
    else:
        USER_LOG.warning(
            "Tags applied with errors. More details in the log file"
        )
        return FAILED_RETURN_CODE


@tags.command(name='remove')
@return_code_manager
@tags_to_context
@verbose_option
@click.pass_context
def remove(ctx):
    """Remove tags from config from deployed resources"""
    output = ctx.obj['output']
    success = ctx.obj['tags_resource'].safe_remove_tags(output)
    if success:
        USER_LOG.info("Tags removed successfully")
        return OK_RETURN_CODE
    else:
        USER_LOG.warning(
            "Tags removed with errors. More details in the log file"
        )
        return FAILED_RETURN_CODE



import copy

from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.connection.rds_connection import RDSConnection
from syndicate.core.helper import unpack_kwargs, dict_keys_to_upper_camel_case
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import validate_params, \
    build_description_obj, validate_known_params
from syndicate.exceptions import ResourceNotFoundError

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()

CLUSTER_REQUIRED_PARAMS = ['engine']

DB_INSTANCE_REQUIRED_PARAMS = [
    'instance_class',
    'engine'
]

DEPLOY_CLUSTER_KNOWN_PARAMS = [
    'availability_zones',
    'backup_retention_period',
    'character_set_name',
    'database_name',
    'd_b_cluster_parameter_group_name',
    'vpc_security_group_ids',
    'db_subnet_group_name',
    'engine',
    'engine_version',
    'port',
    'master_username',
    'master_user_password',
    'option_group_name',
    'preferred_backup_window',
    'preferred_maintenance_window',
    'replication_source_identifier',
    'tags',
    'storage_encrypted',
    'kms_key_id',
    'pre_signed_url',
    'iam_db_auth',
    'backtrack_window',
    'enable_cloudwatch_logs_exports',
    'engine_mode',
    'scaling_configuration',
    'deletion_protection',
    'global_cluster_identifier',
    'enable_http_endpoint',
    'copy_tags_to_snapshot',
    'domain',
    'domain_i_a_m_role_name',
    'enable_global_write_forwarding',
    'd_b_cluster_instance_class',
    'allocated_storage',
    'storage_type',
    'iops',
    'publicly_accessible',
    'auto_minor_version_upgrade',
    'monitoring_interval',
    'monitoring_role_arn',
    'enable_performance_insights',
    'performance_insights_k_m_s_key_id',
    'performance_insights_retention_period',
    'serverless_v2_scaling_configuration',
    'network_type',
    'd_b_system_id',
    'manage_master_user_password',
    'master_user_secret_kms_key_id',
    'source_region'
]

UPDATE_CLUSTER_KNOWN_PARAMS = [
    'new_d_b_cluster_identifier',
    'apply_immediately',
    'backup_retention_period',
    'd_b_cluster_parameter_group_name',
    'vpc_security_group_ids',
    'port',
    'master_user_password',
    'option_group_name',
    'preferred_backup_window',
    'preferred_maintenance_window',
    'iam_db_auth',
    'backtrack_window',
    'cloudwatch_logs_export_configuration',
    'engine_version',
    'allow_major_version_upgrade',
    'd_b_instance_parameter_group_name',
    'domain',
    'domain_i_a_m_role_name',
    'scaling_configuration',
    'deletion_protection',
    'enable_http_endpoint',
    'copy_tags_to_snapshot',
    'enable_global_write_forwarding',
    'd_b_cluster_instance_class',
    'allocated_storage',
    'storage_type',
    'iops',
    'auto_minor_version_upgrade',
    'monitoring_interval',
    'monitoring_role_arn',
    'enable_performance_insights',
    'performance_insights_k_m_s_key_id',
    'performance_insights_retention_period',
    'serverless_v2_scaling_configuration',
    'network_type',
    'manage_master_user_password',
    'rotate_master_user_password',
    'master_user_secret_kms_key_id'
]

DEPLOY_DB_INSTANCE_KNOWN_PARAMS = [
    'database_name',
    'allocated_storage',
    'instance_class',
    'engine',
    'master_username',
    'master_user_password',
    'd_b_security_groups',
    'vpc_security_group_ids',
    'availability_zone',
    'db_subnet_group_name',
    'preferred_maintenance_window',
    'd_b_parameter_group_name',
    'backup_retention_period',
    'preferred_maintenance_window',
    'd_b_parameter_group_name',
    'preferred_backup_window',
    'port',
    'multi_a_z',
    'engine_version',
    'auto_minor_version_upgrade',
    'license_model',
    'iops',
    'option_group_name',
    'character_set_name',
    'nchar_character_set_name',
    'publicly_accessible',
    'tags',
    'cluster_name',
    'storage_type',
    'tde_credential_arn',
    'tde_credential_password',
    'storage_encrypted',
    'kms_key_id',
    'domain',
    'copy_tags_to_snapshot',
    'monitoring_interval',
    'monitoring_role_arn',
    'domain_i_a_m_role_name',
    'promotion_tier',
    'timezone',
    'iam_db_auth',
    'enable_performance_insights',
    'performance_insights_k_m_s_key_id',
    'performance_insights_retention_period',
    'enable_cloudwatch_logs_exports',
    'processor_features',
    'deletion_protection',
    'max_allocated_storage',
    'enable_customer_owned_ip',
    'custom_iam_instance_profile',
    'backup_target',
    'network_type',
    'storage_throughput',
    'manage_master_user_password',
    'master_user_secret_kms_key_id',
    'c_a_certificate_identifier'
]

UPDATE_DB_INSTANCE_KNOWN_PARAMS = [
    'allocated_storage',
    'instance_class',
    'db_subnet_group_name',
    'd_b_security_groups',
    'vpc_security_group_ids',
    'apply_immediately',
    'master_user_password',
    'd_b_parameter_group_name',
    'backup_retention_period',
    'preferred_backup_window',
    'preferred_maintenance_window',
    'multi_a_z',
    'engine_version',
    'allow_major_version_upgrade',
    'auto_minor_version_upgrade',
    'license_model',
    'iops',
    'option_group_name',
    'storage_type',
    'tde_credential_arn',
    'tde_credential_password',
    'c_a_certificate_identifier',
    'domain',
    'copy_tags_to_snapshot',
    'monitoring_interval',
    'd_b_port_number',
    'publicly_accessible',
    'monitoring_role_arn',
    'domain_i_a_m_role_name',
    'promotion_tier',
    'iam_db_auth',
    'enable_performance_insights',
    'performance_insights_k_m_s_key_id',
    'performance_insights_retention_period',
    'cloudwatch_logs_export_configuration',
    'processor_features',
    'use_default_processor_features',
    'deletion_protection',
    'max_allocated_storage',
    'certificate_rotation_restart',
    'replica_mode',
    'enable_customer_owned_ip',
    'aws_backup_recovery_point_arn',
    'automation_mode',
    'resume_full_automation_mode_minutes',
    'network_type',
    'storage_throughput',
    'manage_master_user_password',
    'rotate_master_user_password',
    'master_user_secret_kms_key_id'
]

UPDATE_CLUSTER_NOT_SUPPORTED_KEYS = [
    'resource_type',
    'engine',
    'master_username',
    'database_name',
    'availability_zones',
    'db_subnet_group_name'
]

UPDATE_INSTANCE_NOT_SUPPORTED_KEYS = [
    'engine',
    'cluster_name',
    'resource_type'
]


class RDSDBClusterResource(BaseResource):

    def __init__(self, rds_conn: RDSConnection) -> None:
        self.rds_conn = rds_conn

    def create_db_cluster(self, args: list) -> dict | tuple:
        """ Create RDS cluster in pool in sub processes.

        :type args: list
        """
        return self.create_pool(self._create_db_cluster_from_meta, args, 1)

    @unpack_kwargs
    def _create_db_cluster_from_meta(self, name: str, meta: dict) -> dict:
        """ Create RDS cluster from meta description.

        :type name: str
        :type meta: dict
        """

        validate_params(name, meta, CLUSTER_REQUIRED_PARAMS)

        description = self.rds_conn.get_db_cluster_description(name)
        if description:
            _LOG.warning(f'RDS cluster {name} already exists.')
            return self.describe_db_cluster(
                name=name, meta=meta, description=description)

        params = copy.deepcopy(meta)
        params.pop('resource_type')

        validate_known_params(name, list(params.keys()),
                              DEPLOY_CLUSTER_KNOWN_PARAMS)

        db_subnet_group_name = params.pop('db_subnet_group_name', None)
        iam_db_auth = params.pop('iam_db_auth', None)

        params = dict_keys_to_upper_camel_case(params)

        if db_subnet_group_name is not None:
            params['DBSubnetGroupName'] = db_subnet_group_name

        if iam_db_auth is not None:
            params['EnableIAMDatabaseAuthentication'] = iam_db_auth

        description = self.rds_conn.create_db_cluster(
            name=name,
            params=params
        )

        USER_LOG.info(
            f"Waiting for the DB cluster '{name}' to become available...")
        waiter = self.rds_conn.get_waiter('db_cluster_available')
        waiter.wait(
            DBClusterIdentifier=name
        )

        USER_LOG.info(f'Endpoint: {description["Endpoint"]}')
        USER_LOG.info(
            f'Reader endpoint: {description["ReaderEndpoint"]}')
        if description.get('CustomEndpoints'):
            endpoints = '\n'.join(description['CustomEndpoints'])
            USER_LOG.info(f'Custom endpoints: {endpoints}')

        return self.describe_db_cluster(
            name=name,
            meta=meta,
            description=description
        )

    def update_db_cluster(self, args: list) -> dict | tuple:
        return self.create_pool(self._update_db_cluster, args, 1)

    @unpack_kwargs
    def _update_db_cluster(self, name: str, meta: dict, context) -> dict:
        cluster = self.rds_conn.get_db_cluster_description(name)
        if not cluster:
            raise ResourceNotFoundError(
                f"RDS DB cluster '{name}' does not exist."
            )

        params = copy.deepcopy(meta)

        # TODO implement tags updating
        params.pop('tags', None)

        for key in UPDATE_CLUSTER_NOT_SUPPORTED_KEYS:
            params.pop(key, None)

        validate_known_params(name, list(params.keys()),
                              UPDATE_CLUSTER_KNOWN_PARAMS)

        db_subnet_group_name = params.pop('db_subnet_group_name', None)
        iam_db_auth = params.pop('iam_db_auth', None)

        params = dict_keys_to_upper_camel_case(params)

        if db_subnet_group_name is not None:
            params['DBSubnetGroupName'] = db_subnet_group_name

        if iam_db_auth is not None:
            params['EnableIAMDatabaseAuthentication'] = iam_db_auth

        params['DBClusterIdentifier'] = name

        description = self.rds_conn.update_db_cluster(
            params=params
        )

        USER_LOG.info(f'Endpoint: {description["Endpoint"]}')
        USER_LOG.info(
            f'Reader endpoint: {description["ReaderEndpoint"]}')
        if description.get('CustomEndpoints'):
            endpoints = '\n'.join(description['CustomEndpoints'])
            USER_LOG.info(f'Custom endpoints: {endpoints}')

        _LOG.info(f"RDS DB cluster '{name}' updated successfully")

        return self.describe_db_cluster(
            name=name,
            meta=meta,
            description=description
        )

    def describe_db_cluster(self, name: str, meta: dict,
                            description: dict = None) -> dict:
        if not description:
            description = self.rds_conn.get_db_cluster_description(name)
            if not description:
                return {}

        arn = description['DBClusterArn']

        return {
            arn: build_description_obj(description, name, meta)
        }

    def remove_db_cluster(self, args: list) -> dict | tuple:
        return self.create_pool(self._remove_db_cluster, args)

    @unpack_kwargs
    def _remove_db_cluster(self, arn: str, config: dict) -> dict:
        from click import confirm as click_confirm

        cluster_name = config['resource_name']
        USER_LOG.info(f"Deleting RDS DB cluster '{cluster_name}'")
        description = \
            self.rds_conn.get_db_cluster_description(cluster_name)

        if not description:
            _LOG.warning(f'RDS DB cluster {cluster_name} is not found')
            return {arn: config}

        if description.get('DeletionProtection') is True:
            if not click_confirm(
                    f"The '{cluster_name}' has deletion protection "
                    f"enabled. Do you want to proceed with deletion?"):
                USER_LOG.info(
                    f"RDS DB cluster '{cluster_name}' deletion skipped")
                return {}

            self._disable_db_cluster_del_protection(cluster_name)

        self.rds_conn.delete_db_cluster(cluster_name)

        USER_LOG.info(f"Waiting for DB cluster '{cluster_name}' deletion...")
        waiter = self.rds_conn.get_waiter('db_cluster_deleted')
        waiter.wait(DBClusterIdentifier=cluster_name)
        _LOG.info(f"RDS DB cluster '{cluster_name}' was removed.")

        return {arn: config}

    def _disable_db_cluster_del_protection(self, name: str):
        USER_LOG.info(
            'Disabling DB cluster deletion protection. Please wait.'
        )
        self.rds_conn.update_db_cluster({
            'DBClusterIdentifier': name,
            'DeletionProtection': False
        })

        USER_LOG.info(
            f"Waiting for the DB cluster '{name}' to become available...")
        waiter = self.rds_conn.get_waiter('db_cluster_available')
        waiter.wait(
            DBClusterIdentifier=name
        )


class RDSDBInstanceResource(BaseResource):

    def __init__(self, rds_conn: RDSConnection) -> None:
        self.rds_conn = rds_conn

    def create_db_instance(self, args: list) -> dict | tuple:
        """ Create RDS db instance in pool in sub processes.

        :type args: list
        """
        return self.create_pool(self._create_db_instance_from_meta, args, 1)

    @unpack_kwargs
    def _create_db_instance_from_meta(self, name: str, meta: dict) -> dict:
        """ Create RDS db instance from meta description.

        :type name: str
        :type meta: dict
        """
        validate_params(name, meta, DB_INSTANCE_REQUIRED_PARAMS)

        cluster_name = meta.pop('cluster_name', None)

        description = self.rds_conn.get_db_instance_description(
            cluster_name=cluster_name,
            instance_name=name
        )
        if description:
            _LOG.warning(f'RDS DB instance {name} already exists.')
            return self.describe_db_instance(
                name=name, meta=meta, description=description)

        params = copy.deepcopy(meta)
        params.pop('resource_type')

        validate_known_params(name, list(params.keys()),
                              DEPLOY_DB_INSTANCE_KNOWN_PARAMS)

        instance_class = params.pop('instance_class', None)
        database_name = params.pop('database_name', None)
        iam_db_auth = params.pop('iam_db_auth', None)
        db_subnet_group_name = params.pop('db_subnet_group_name', None)

        params = dict_keys_to_upper_camel_case(params)

        if cluster_name is not None:
            params['DBClusterIdentifier'] = cluster_name

        if instance_class is not None:
            params['DBInstanceClass'] = instance_class

        if database_name is not None:
            params['DBName'] = database_name

        if iam_db_auth is not None:
            params['EnableIAMDatabaseAuthentication'] = iam_db_auth

        if db_subnet_group_name is not None:
            params['DBSubnetGroupName'] = db_subnet_group_name

        description = self.rds_conn.create_db_instance(
            name=name,
            params=params
        )

        USER_LOG.info(
            f"Waiting for the DB instance '{name}' to become available. This "
            f"may take up to 15 minutes. Please refrain from interrupting."
        )
        waiter = self.rds_conn.get_waiter('db_instance_available')
        waiter.wait(
            DBInstanceIdentifier=name
        )

        return self.describe_db_instance(
            name=name,
            meta=meta,
            description=description
        )

    def update_db_instance(self, args: list) -> dict | tuple:
        return self.create_pool(self._update_db_instance, args, 1)

    @unpack_kwargs
    def _update_db_instance(self, name: str, meta: dict, context) -> dict:
        instance = self.rds_conn.get_db_instance_description(
            cluster_name=meta.get('cluster_name'),
            instance_name=name
        )
        if not instance:
            raise ResourceNotFoundError(
                f"RDS DB instance '{name}' does not exist."
            )

        params = copy.deepcopy(meta)

        # TODO implement tags updating
        params.pop('tags', None)

        for key in UPDATE_INSTANCE_NOT_SUPPORTED_KEYS:
            params.pop(key, None)

        validate_known_params(
            name,
            list(params.keys()),
            UPDATE_DB_INSTANCE_KNOWN_PARAMS
        )

        instance_class = params.pop('instance_class', None)
        database_name = params.pop('database_name', None)
        iam_db_auth = params.pop('iam_db_auth', None)
        db_subnet_group_name = params.pop('db_subnet_group_name', None)

        params = dict_keys_to_upper_camel_case(params)

        params['DBInstanceIdentifier'] = name

        if instance_class is not None:
            params['DBInstanceClass'] = instance_class

        if database_name is not None:
            params['DBName'] = database_name

        if iam_db_auth is not None:
            params['EnableIAMDatabaseAuthentication'] = iam_db_auth

        if db_subnet_group_name is not None:
            params['DBSubnetGroupName'] = db_subnet_group_name

        _LOG.info('Updating db instance...')
        description = self.rds_conn.update_db_instance(
            params=params
        )

        USER_LOG.info(
            f"Waiting for DB instance '{name}' to become available...")
        waiter = self.rds_conn.get_waiter('db_instance_available')
        waiter.wait(DBInstanceIdentifier=name)

        _LOG.info(f"RDS DB instance '{name}' updated successfully")

        return self.describe_db_instance(
            name=name,
            meta=meta,
            description=description
        )

    def describe_db_instance(self, name: str, meta: dict,
                             description: dict = None) -> dict:
        if not description:
            description = self.rds_conn.get_db_instance_description(
                cluster_name=meta.get('cluster_name'),
                instance_name=name
            )
            if not description:
                return {}

        arn = description['DBInstanceArn']

        return {
            arn: build_description_obj(description, name, meta)
        }

    def remove_db_instance(self, args: list) -> dict | tuple:
        return self.create_pool(self._remove_db_instance, args)

    @unpack_kwargs
    def _remove_db_instance(self, arn: str, config: dict) -> dict:
        from click import confirm as click_confirm

        instance_name = config['resource_name']
        db_cluster_name = \
            config['resource_meta'].get('cluster_name', '')

        instance_description = \
            self.rds_conn.get_db_instance_description(
                cluster_name=db_cluster_name,
                instance_name=instance_name)

        if not instance_description:
            _LOG.warning(f'RDS DB instance {instance_name} is not found')
            return {arn: config}

        USER_LOG.info(f"Deleting RDS DB instance '{instance_name}'")

        cluster_del_protection = \
            self.rds_conn.is_db_cluster_del_protection(db_cluster_name)
        if cluster_del_protection:
            if not click_confirm(
                    f"The '{instance_name}' DB cluster has deletion "
                    f"protection enabled. Do you want to proceed with "
                    f"deletion?"):
                return {}

            self.rds_conn.disable_db_cluster_del_protection(db_cluster_name)

        if instance_description.get('DeletionProtection') is True:
            if not click_confirm(
                    f"The '{instance_name}' has deletion protection "
                    f"enabled. Do you want to proceed with deletion?"):
                return {}

            self.rds_conn.disable_db_instance_del_protection(instance_name)

        self.rds_conn.delete_db_instance(instance_name)

        USER_LOG.info(
            f"Waiting for DB instance '{instance_name}' deletion. This may "
            f"take up to 15 minutes. Please refrain from interrupting.")
        waiter = self.rds_conn.get_waiter('db_instance_deleted')
        waiter.wait(DBInstanceIdentifier=instance_name)
        _LOG.info(f"RDS DB instance '{instance_name}' was removed.")

        if cluster_del_protection:
            self.rds_conn.enable_db_cluster_del_protection(db_cluster_name)

        return {arn: config}



"""
    Copyright 2024 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import concurrent
import json
import os

from concurrent.futures.thread import ThreadPoolExecutor
from pathlib import Path

from syndicate.exceptions import ArtifactAssemblingError, \
    EnvironmentError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.build.helper import build_py_package_name, zip_dir, \
    remove_dir, run_external_command

from syndicate.core.constants import LAMBDA_CONFIG_FILE_NAME, \
    LAMBDA_LAYER_CONFIG_FILE_NAME
from syndicate.core.helper import unpack_kwargs, build_path
from syndicate.core.resources.helper import validate_params


BIN_DIR = 'bin'
OBJ_DIR = 'obj'
X64_DIR = 'x64'
NET_8_0_DIR = 'net8.0'
DOTNET_CORE_DIR = 'dotnetcore'
STORE_DIR = 'store'

BUILD_DIR_TMP = 'dotnet_tmp'
LAMBDA_DIR = 'lambda'
LAYER_DIR = 'layer'

ARTIFACT_FILE = 'artifact.xml'

BUILD_LAMBDA_TMP_DIRS = [BIN_DIR, OBJ_DIR]
BUILD_LAYER_TMP_DIRS = [OBJ_DIR]

CHECK_DOTNET_INSTALLED_COMMAND = ['dotnet', '--info']

SYNDICATE_DIR = '.syndicate'
LOCAL_NUGET_SOURCE_NAME = 'syndicate_local_nuget_source'


_LOG = get_logger(__name__)
USER_LOG = get_user_logger()


def assemble_dotnet_lambdas(project_path, bundles_dir, **kwargs):
    from syndicate.core import CONFIG

    _check_dotnet_is_installed()
    project_abs_path = Path(CONFIG.project_path, project_path)
    _LOG.info(f'Going to package lambdas starting by path {project_abs_path}')
    executor = ThreadPoolExecutor(max_workers=5)
    futures = []
    for root, sub_dirs, files in os.walk(project_abs_path):
        for item in files:
            if item.endswith(LAMBDA_LAYER_CONFIG_FILE_NAME):
                _LOG.info(f'Going to build artifact in: {root}')
                arg = {
                    'item': item,
                    'root': root,
                    'target_folder': bundles_dir
                }
                futures.append(executor.submit(
                    _build_dotnet_lambda_layer_artifact, arg))
    for future in concurrent.futures.as_completed(futures):
        if future.result():
            _LOG.info(future.result())

    futures = []
    for root, sub_dirs, files in os.walk(project_abs_path):
        for item in files:
            if item.endswith(LAMBDA_CONFIG_FILE_NAME):
                _LOG.info(f'Going to build artifact in: {root}')
                arg = {
                    'item': item,
                    'root': root,
                    'target_folder': bundles_dir
                }
                futures.append(executor.submit(
                    _build_dotnet_lambda_artifact, arg))
    for future in concurrent.futures.as_completed(futures):
        if future.result():
            _LOG.info(future.result())

    _clean_tmp_files(bundles_dir, [BUILD_DIR_TMP])


@unpack_kwargs
def _build_dotnet_lambda_artifact(item, root, target_folder):
    _LOG.debug(f'Building artifact in {target_folder}')
    lambda_config_dict = json.load(open(build_path(root, item)))
    _LOG.debug(f'Root path: {root}')
    req_params = ['lambda_path', 'name', 'version']
    validate_params(root, lambda_config_dict, req_params)
    lambda_name = lambda_config_dict['name']
    lambda_version = lambda_config_dict['version']
    package_name = build_py_package_name(lambda_name, lambda_version)
    layers = lambda_config_dict.get('layers', [])
    output_path = build_path(target_folder, BUILD_DIR_TMP, LAMBDA_DIR,
                             lambda_name)

    command = [
        'dotnet', 'publish', root,
        '-p:GenerateRuntimeConfigurationFiles=true',
        '--configuration', 'Release',
        '--framework', 'net8.0',
        '--runtime', 'linux-x64',
        '--self-contained', 'False',
        '--output', output_path
    ]

    for layer_name in layers:
        layer_artifact_path = build_path(target_folder, BUILD_DIR_TMP,
                                         LAYER_DIR, layer_name,
                                         DOTNET_CORE_DIR, STORE_DIR, X64_DIR,
                                         NET_8_0_DIR, ARTIFACT_FILE)
        command.extend(['--manifest', layer_artifact_path])
        if not Path(layer_artifact_path).is_file():
            USER_LOG.warn(f'The \'{ARTIFACT_FILE}\' file for the layer '
                          f'\'{layer_name}\' is not found, it might lead to '
                          f'lambda \'{lambda_name}\' artifact generation '
                          f'failure.')

    exit_code, stdout, stderr = run_external_command(command)
    if exit_code != 0:
        raise ArtifactAssemblingError(
            f"An error occurred during lambda '{lambda_name}' packaging. "
            f"Details:\n{stdout or ''}\n{stderr or ''}"
        )
    _LOG.info(f'Running the command "{command}"\n{stdout or ""}'
              f'\n{stderr or ""}')

    zip_dir(output_path, build_path(target_folder, package_name))

    _clean_tmp_files(root, BUILD_LAMBDA_TMP_DIRS)


@unpack_kwargs
def _build_dotnet_lambda_layer_artifact(item, root, target_folder):
    _LOG.debug(f'Building artifact in {target_folder}')
    layer_config = json.load(open(build_path(root, item)))
    _LOG.debug(f'Root path: {root}')

    layer_name = layer_config['name']
    package_name = layer_config['deployment_package']
    custom_packages = layer_config.get('custom_packages', [])
    validate_params(root, layer_config, ['name', 'deployment_package'])

    output_path = build_path(target_folder, BUILD_DIR_TMP, LAYER_DIR,
                             layer_name, DOTNET_CORE_DIR, STORE_DIR)

    if custom_packages:
        _LOG.info(f'Processing layer \'{layer_name}\' custom packages '
                  f'\'{custom_packages}\'.')
        _process_custom_packages(root, custom_packages)

    _LOG.info(f'Packaging artifacts {package_name}')
    command = [
        'dotnet', 'store', '--skip-optimization',
        '--manifest', root,
        '--framework', 'net8.0',
        '--runtime', 'linux-x64',
        '--output', output_path
    ]

    exit_code, stdout, stderr = run_external_command(command)
    if exit_code != 0:
        raise ArtifactAssemblingError(
            f"An error occurred during lambda layer '{layer_name}' "
            f"packaging. Details:\n{stdout or ''}\n{stderr or ''}"
        )
    _LOG.info(f'Running the command "{command}"\n{stdout or ""}'
              f'\n{stderr or ""}')

    zip_dir(
        build_path(target_folder, BUILD_DIR_TMP, LAYER_DIR, layer_name),
        build_path(target_folder, package_name))

    _clean_tmp_files(root, BUILD_LAYER_TMP_DIRS)


def _check_dotnet_is_installed():
    try:
        exit_code, _, _ = run_external_command(CHECK_DOTNET_INSTALLED_COMMAND)
    except Exception as e:
        raise EnvironmentError(
            'It seems like the dotnet SDK is not installed. There is no '
            'ability to build a DotNet bundle. Please, make sure dotnet SDK '
            'is installed and retry to build a bundle.')


def _clean_tmp_files(location, dirs):
    for tmp_dir in dirs:
        _LOG.debug(f'Remove tmp directory {tmp_dir}')
        remove_dir(build_path(location, tmp_dir))


def _process_custom_packages(layer_dir, packages):
    if not _is_local_source_exist(LOCAL_NUGET_SOURCE_NAME):
        _LOG.info(
            'Syndicate local NuGet source does not exist, creating...')
        _create_local_nuget_source()

    for package in packages:
        _LOG.info(f'Publishing package \'{package}\' to local NuGet source '
                  f'\'{LOCAL_NUGET_SOURCE_NAME}\'...')

        package_path = build_path(layer_dir, package)
        command = [
            'dotnet', 'nuget',
            'push', package_path,
            '--source', LOCAL_NUGET_SOURCE_NAME
        ]

        exit_code, stdout, stderr = run_external_command(command)
        if exit_code != 0:
            raise EnvironmentError(
                f'An error occurred during publishing package \'{package}\' '
                f'into local NuGet source \'{LOCAL_NUGET_SOURCE_NAME}\'.'
                f'Details:\n{stdout or ""}\n{stderr or ""}'
            )
    _LOG.info(f'All custom packages successfully published to local NuGet '
              f'source \'{LOCAL_NUGET_SOURCE_NAME}\'.')


def _create_local_nuget_source():
    home_dir = str(Path.home())
    local_store = build_path(home_dir, SYNDICATE_DIR, LOCAL_NUGET_SOURCE_NAME)
    if not Path(local_store).exists():
        _LOG.info('Directory for local NuGet source does not exist, '
                  'creating...')
        Path(local_store).mkdir(parents=True)

    command = [
        'dotnet', 'nuget', 'add', 'source', local_store,
        '--name', LOCAL_NUGET_SOURCE_NAME
    ]

    exit_code, stdout, stderr = run_external_command(command)
    if exit_code != 0:
        raise EnvironmentError(
            f'An error occurred during local NuGet source creation.'
            f'Details:\n{stdout or ""}\n{stderr or ""}'
        )

    _LOG.info(f'Syndicate local NuGet source \'{LOCAL_NUGET_SOURCE_NAME}\' '
              f'created successfully.')


def _is_local_source_exist(source_name):
    command = ['dotnet', 'nuget', 'list', 'source', '--format', 'Short']
    exit_code, stdout, stderr = run_external_command(command)
    if exit_code != 0:
        raise EnvironmentError(
            f'An error occurred during an attempt to list NuGet sources.'
            f'Details:\n{stdout or ""}\n{stderr or ""}'
        )
    return True if source_name in stdout else False



from boto3 import client
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger(__name__)


@apply_methods_decorator(retry())
class EventBridgeSchedulerConnection(object):
    """ EventBridge Scheduler Connection class."""

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.client = client('scheduler', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new EventBridge Scheduler connection.')

    def create_schedule(self, **params):
        return self.client.create_schedule(**params)['ScheduleArn']

    def update_schedule(self, **params):
        return self.client.update_schedule(**params)['ScheduleArn']

    def describe_schedule(self, name, group_name=None):
        params = {'Name': name}
        if group_name is not None:
            params['GroupName'] = group_name

        try:
            return self.client.get_schedule(**params)

        except ClientError as e:
            if 'ResourceNotFoundException' in str(e):
                _LOG.warning(
                    f'Cannot find EventBridge schedule with name {name}')
                return None
            else:
                raise e

    def delete_schedule(self, name, group_name=None, log_not_found_error=True):
        """
        log_not_found_error parameter is needed for proper log handling in the
        retry decorator
        """
        params = {'Name': name}
        if group_name is not None:
            params['GroupName'] = group_name

        return self.client.delete_schedule(**params)


"""
    Copyright 2024 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json
import os
import shutil
import zipfile

from syndicate.exceptions import ArtifactAssemblingError, \
    ResourceMetadataError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.constants import APPSYNC_ARTIFACT_NAME_TEMPLATE, \
    APPSYNC_CONFIG_FILE_NAME
from syndicate.core.helper import build_path

FILE_DEPLOYMENT_RESOURCES = 'deployment_resources.json'

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()


def assemble_appsync(project_path, bundles_dir, **kwargs):
    from syndicate.core import CONFIG
    path_to_project = CONFIG.project_path
    src_path = build_path(path_to_project, project_path)
    if not os.path.exists(src_path):
        raise ArtifactAssemblingError(
            f"Appsync sources are not located by path '{src_path}'")
    if not os.listdir(src_path):
        raise ArtifactAssemblingError(
            f"Appsync sources path '{src_path}' is empty"
        )

    _LOG.info(f'Appsync sources are located by path: {src_path}')

    for item in os.listdir(src_path):
        appsync_src_path = build_path(src_path, item)
        if os.path.isdir(appsync_src_path):
            _LOG.info(f'Going to process \'{item}\'')
            if not os.listdir(appsync_src_path):
                raise ArtifactAssemblingError(
                    f"Appsync path '{appsync_src_path}' is empty"
                )

            conf_file_path = build_path(src_path, item,
                                        APPSYNC_CONFIG_FILE_NAME)
            if not os.path.isfile(conf_file_path):
                raise ResourceMetadataError(
                    f'\'{APPSYNC_CONFIG_FILE_NAME}\' file not found for '
                    f'Appsync \'{item}\'.')

            with open(conf_file_path) as file:
                appsync_conf = json.load(file)
            schema_filepath = appsync_conf.get('schema_path')
            if not os.path.isabs(schema_filepath):
                schema_path = build_path(appsync_src_path, schema_filepath)
                USER_LOG.info(f'Path to schema file resolved as '
                              f'\'{schema_path}\'')
            else:
                schema_path = schema_filepath
            if not os.path.isfile(schema_path):
                raise ArtifactAssemblingError(
                    f'Schema file not found for Appsync \'{item}\' in the '
                    f'path {schema_path}.')

            artifact_name = APPSYNC_ARTIFACT_NAME_TEMPLATE.format(name=item)
            zip_file_path = build_path(src_path, artifact_name)
            appsync_conf['deployment_package'] = artifact_name

            with open(conf_file_path, 'w') as file:
                json.dump(appsync_conf, file)

            resolvers = appsync_conf.get('resolvers', [])
            functions = appsync_conf.get('functions', [])
            artifact_items = resolvers + functions
            artifacts_path = []
            for artifact_item in artifact_items:
                if artifact_file_path := artifact_item.get(
                        'request_mapping_template_path'):
                    artifacts_path.append(artifact_file_path)
                if artifact_file_path := artifact_item.get(
                        'response_mapping_template_path'):
                    artifacts_path.append(artifact_file_path)
                if artifact_file_path := artifact_item.get('code_path'):
                    artifacts_path.append(artifact_file_path)

            with zipfile.ZipFile(zip_file_path, 'w') as zipf:
                zipf.write(schema_path, schema_filepath)

                # to archive only resolvers and functions in config
                for path in artifacts_path:
                    artifacts_path = build_path(appsync_src_path, path)
                    if os.path.exists(artifacts_path):
                        zipf.write(artifacts_path, path)
                        _LOG.debug(
                            f"File from the path '{artifacts_path}' "
                            f"successfully added to the AppSync deployment "
                            f"package")
                    else:
                        raise ArtifactAssemblingError(
                            f"File not found by the path '{artifacts_path}'")

            if os.path.getsize(zip_file_path) == 0:
                raise ArtifactAssemblingError(
                    f'Appsync archive {zip_file_path} is empty')

            shutil.move(zip_file_path, build_path(bundles_dir, artifact_name))
            _LOG.info(f'Appsync {item} was processed successfully')



from syndicate.core.constants import BATCH_COMPENV_TYPE
from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator


class BatchCompenvGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = BATCH_COMPENV_TYPE
    CONFIGURATION = {
        "compute_environment_type": "MANAGED",
        "state": "ENABLED",
        "service_role": None,
        "compute_resources": {  # the ones from boto3 just in snake case
            "type": "EC2",
            "allocation_strategy": None,
            "minv_cpus": 0,
            "maxv_cpus": 8,
            "desiredv_cpus": 1,
            "instance_types": list,
            "security_group_ids": list,
            "subnets": list,
            "instance_role": None,
        },
        "tags": dict
    }

    def __init__(self, **kwargs):
        if kwargs.get('type') != 'FARGATE' and not kwargs['instance_types']:
            kwargs['instance_types'] = ['optimal']
        super().__init__(**kwargs)

    def _generate_resource_configuration(self) -> dict:
        result = super()._generate_resource_configuration()
        if result['compute_resources']['type'] in ['FARGATE', 'FARGATE_SPOT']:
            result['compute_resources'].pop('desiredv_cpus')
            result['compute_resources'].pop('minv_cpus')
            # Should I remove 'instance_role' here? because docs and
            # 'validate_batch_compenv' urges me, but aws api thinks differently
        return result



import io
import os.path
import posixpath
import shutil
import time
from pathlib import PurePath
from zipfile import ZipFile

from botocore.exceptions import ClientError

from syndicate.exceptions import ArtifactError, \
    ResourceProcessingError, ResourceNotFoundError, ParameterError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.constants import ARTIFACTS_FOLDER
from syndicate.core.helper import build_path, unpack_kwargs, \
    dict_keys_to_camel_case
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import validate_params, \
    build_description_obj

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()

API_REQUIRED_PARAMS = ['schema_path']
DATA_SOURCE_REQUIRED_PARAMS = ['name', 'type']
RESOLVER_REQUIRED_PARAMS = ['type_name', 'field_name', 'runtime']
RESOLVER_DEFAULT_KIND = 'UNIT'

DEFAULT_FUNC_VTL_VERSION = '2018-05-29'
FUNCTION_REQUIRED_PARAMS = ['runtime', 'data_source_name']

AWS_REGION_PARAMETER = 'aws_region'
AWS_LAMBDA_TYPE = 'AWS_LAMBDA'
AWS_CUP_TYPE = 'AMAZON_COGNITO_USER_POOLS'

DATA_SOURCE_TYPE_CONFIG_MAPPING = {
    AWS_LAMBDA_TYPE: 'lambda_config',
    'AMAZON_DYNAMODB': 'dynamodb_config',
    'AMAZON_ELASTICSEARCH': 'elasticsearch_config',
    'HTTP': 'http_config',
    'RELATIONAL_DATABASE': 'relational_database_config',
    'AMAZON_OPENSEARCH_SERVICE': 'open_search_service_config',
    'AMAZON_EVENTBRIDGE': 'event_bridge_config'
}

ONE_DAY_IN_SECONDS = 86400


class AppSyncResource(BaseResource):

    def __init__(self, appsync_conn, s3_conn, cup_conn, cw_logs_conn,
                 deploy_target_bucket, deploy_target_bucket_key_compound,
                 account_id) -> None:
        from syndicate.core import CONF_PATH
        self.appsync_conn = appsync_conn
        self.s3_conn = s3_conn
        self.cup_conn = cup_conn
        self.cw_logs_conn = cw_logs_conn
        self.conf_path = CONF_PATH
        self.deploy_target_bucket_key_compound = \
            deploy_target_bucket_key_compound
        self.deploy_target_bucket = deploy_target_bucket
        self.account_id = account_id

    def create_graphql_api(self, args):
        """ Create GraphQL API in pool in sub processes.

        :type args: list
        """
        return self.create_pool(self._create_graphql_api_from_meta, args, 1)

    @unpack_kwargs
    def _create_graphql_api_from_meta(self, name, meta):
        """ Create GraphQL API from meta description.

        :type name: str
        :type meta: dict
        """
        extract_to = None
        validate_params(name, meta, API_REQUIRED_PARAMS)

        api = self.appsync_conn.get_graphql_api_by_name(name)
        if api:
            _LOG.warning(f'AppSync API {name} already exists.')
            return self.describe_graphql_api(
                name=name, meta=meta, api_id=api['apiId'])

        archive_path = meta.get('deployment_package')
        if archive_path:
            extract_to = self._extract_zip(archive_path, name)
        auth_type = meta.get('primary_auth_type')
        extra_auth_types = meta.get('extra_auth_types', [])
        lambda_auth_config = meta.get('lambda_authorizer_config', {})
        user_pool_config = meta.get('user_pool_config', {})

        if auth_type == AWS_LAMBDA_TYPE:
            lambda_region = lambda_auth_config.pop(AWS_REGION_PARAMETER, None)
            lambda_name = lambda_auth_config.pop('resource_name', None)
            if not lambda_region or not lambda_name:
                lambda_auth_config = {}
                USER_LOG.error(
                    f"Authorization type '{auth_type}' can't be configured "
                    f"for AppSync '{name}' because lambda resource name or "
                    f"aws region is not specified")
            else:
                lambda_arn = self.build_lambda_arn(
                    region=lambda_region,
                    name=lambda_name)
                lambda_auth_config['authorizer_uri'] = lambda_arn

        if auth_type == AWS_CUP_TYPE:
            cup_name = user_pool_config.pop('resource_name', None)
            cup_id = self.cup_conn.if_pool_exists_by_name(cup_name)
            if cup_id:
                user_pool_config['user_pool_id'] = cup_id
            else:
                user_pool_config = {}
                USER_LOG.error(
                    f"Authorization type '{auth_type}' can't be configured "
                    f"for AppSync '{name}' because Cognito User Pool "
                    f"{cup_name} not found")

        updated_extra_auth_types, is_extra_auth_api_key = \
            self._process_extra_auth(extra_auth_types, name)

        api_id = self.appsync_conn.create_graphql_api(
            name, auth_type=auth_type, tags=meta.get('tags'),
            user_pool_config=user_pool_config,
            open_id_config=meta.get('open_id_config'),
            lambda_auth_config=lambda_auth_config,
            log_config=self._resolve_log_config(
                meta.get('log_config', {}), name
            ),
            xray_enabled=meta.get('xray_enabled'),
            extra_auth_types=updated_extra_auth_types)

        if auth_type == 'API_KEY' or is_extra_auth_api_key:
            api_key_expiration = meta.get('api_key_expiration_days', 7)
            now = time.time()
            api_key_expires = \
                int(now + api_key_expiration * ONE_DAY_IN_SECONDS)
            self.appsync_conn.create_api_key(api_id, expires=api_key_expires)

        if schema_path := meta.get('schema_path'):
            schema_full_path = build_path(extract_to, schema_path)
            if not extract_to or not os.path.exists(schema_full_path):
                raise ArtifactError(
                    f'\'{schema_full_path}\' file not found for '
                    f'AppSync \'{name}\''
                )

            with open(schema_full_path, 'r', encoding='utf-8') as file:
                schema_definition = file.read()

            status, details = \
                self.appsync_conn.create_schema(api_id, schema_definition)
            if status != 'SUCCESS':
                error_message = (
                    f"An error occurred when creating schema. "
                    f"Operation status: '{status}'. ")
                if details:
                    error_message += f"Details: '{details}'"
                raise ResourceProcessingError(error_message)
            else:
                _LOG.info(
                    f"Schema of the AppSync '{name}' created successfully")

        data_sources_meta = meta.get('data_sources', [])
        for data_source_meta in data_sources_meta:
            params = self._build_data_source_params_from_meta(data_source_meta)
            if not params:
                continue
            self.appsync_conn.create_data_source(api_id, **params)

        functions_config = []
        functions_meta = meta.get('functions', [])
        for func_meta in functions_meta:
            params = self._build_function_params_from_meta(
                func_meta, extract_to)
            if not params:
                continue
            func_config = self.appsync_conn.create_function(api_id, params)
            functions_config.append(func_config)

        resolvers_meta = meta.get('resolvers', [])
        for resolver_meta in resolvers_meta:
            params = self._build_resolver_params_from_meta(
                resolver_meta, extract_to, functions_config)
            if not params:
                continue
            self.appsync_conn.create_resolver(api_id, **params)

        if extract_to:
            shutil.rmtree(extract_to, ignore_errors=True)
        _LOG.info(f'Created AppSync GraphQL API {api_id}')
        return self.describe_graphql_api(name=name, meta=meta, api_id=api_id)

    def _build_data_source_params_from_meta(self, source_meta: dict):
        source_name = source_meta.get('name')
        try:
            validate_params(
                source_name, source_meta, DATA_SOURCE_REQUIRED_PARAMS)
        except ParameterError as e:
            _LOG.warning(str(e))
            _LOG.warning(f'Skipping data source \'{source_name}\'...')
            return

        _LOG.info(f'Altering data source \'{source_name}\'...')
        source_config = None
        source_type = source_meta.get('type')

        data_source_params = {
            'name': source_name,
            'source_type': source_type
        }

        if source_description := source_meta.get('description'):
            data_source_params['description'] = source_description

        if role_name := source_meta.get('service_role_name'):
            role_arn = self._build_iam_role_arn(role_name)
            data_source_params['service_role_arn'] = role_arn

        if config_key := DATA_SOURCE_TYPE_CONFIG_MAPPING.get(source_type):
            source_config = source_meta.get(config_key)

        if source_type == AWS_LAMBDA_TYPE and source_config:
            region = source_config.pop(AWS_REGION_PARAMETER, None)
            lambda_name = source_config.pop('lambda_name', None)
            source_config['lambda_function_arn'] = self.build_lambda_arn(
                lambda_name, region)
        elif source_type == 'AMAZON_EVENTBRIDGE' and source_config:
            region = source_config.pop(AWS_REGION_PARAMETER)
            event_bus = source_config.pop('event_bus_name')
            source_config['event_bus_arn'] = self.build_event_bus_arn(
                event_bus, region)

        if source_config:
            data_source_params['source_config'] = source_config

        return data_source_params

    def build_lambda_arn(self, name, region):
        arn = f'arn:aws:lambda:{region}:{self.account_id}:function:{name}'
        return arn

    def build_event_bus_arn(self, name, region):
        arn = f'arn:aws:events:{region}:{self.account_id}:event-bus/{name}'
        return arn

    def _build_iam_role_arn(self, name):
        return f'arn:aws:iam::{self.account_id}:role/{name}'

    @staticmethod
    def _build_resolver_params_from_meta(resolver_meta, artifacts_path,
                                         existing_functions=None):
        existing_functions = existing_functions or []
        type_name = resolver_meta.get('type_name')
        field_name = resolver_meta.get('field_name')
        try:
            validate_params(type_name + ':' + field_name, resolver_meta,
                            RESOLVER_REQUIRED_PARAMS)
        except ParameterError as e:
            _LOG.warning(str(e))
            _LOG.warning(f'Skipping resolver for type \'{type_name}\' '
                         f'and field \'{field_name}\'...')
            return

        _LOG.info(f'Altering resolver for type \'{type_name}\' and field '
                  f'\'{field_name}\'...')
        resolver_params = {
            'type_name': type_name,
            'field_name': field_name,
        }
        kind = resolver_meta.get('kind', RESOLVER_DEFAULT_KIND)
        resolver_params['kind'] = kind
        if kind == RESOLVER_DEFAULT_KIND:
            resolver_params['data_source_name'] = \
                resolver_meta['data_source_name']
        elif kind == 'PIPELINE':
            _LOG.info("Trying to resolve functions IDs")
            function_names = \
                resolver_meta.get('pipeline_config', {}).get('functions', [])
            function_ids = []
            for func_name in function_names:
                func_id = None
                for func_info in existing_functions:
                    if func_name == func_info['name']:
                        func_id = func_info['functionId']
                        function_ids.append(func_id)
                        _LOG.debug(
                            f"Successfully resolved ID '{func_id}' for the "
                            f"function '{func_name}'")
                        break
                if not func_id:
                    _LOG.warning(f"ID for the function '{func_name}' was "
                                 f"not resolved")

            resolver_params['pipeline_config'] = {
                'functions': function_ids
            }

        if resolver_meta.get('runtime') in ('JS', 'APPSYNC_JS'):
            runtime = {
                'name': 'APPSYNC_JS',
                'runtimeVersion': '1.0.0'
            }
            resolver_params['runtime'] = runtime
        else:
            runtime = None

        if runtime:
            code_path = build_path(artifacts_path,
                                   resolver_meta.get('code_path'))
            if not artifacts_path or not os.path.exists(code_path):
                raise ArtifactError(
                    f"Resolver code file for type '{type_name}' and field "
                    f"'{field_name}' not found.")

            with open(code_path, 'r', encoding='utf-8') as file:
                code = file.read()
            resolver_params['code'] = code
        else:
            _LOG.debug('Runtime is not JS')
            request_template_path = build_path(
                artifacts_path,
                resolver_meta.get('request_mapping_template_path'))
            if not artifacts_path or not os.path.exists(request_template_path):
                raise ArtifactError(
                    f"Resolver request mapping template file for type "
                    f"'{type_name}' and field '{field_name}' not found.")
            else:
                with open(request_template_path, 'r',
                          encoding='utf-8') as file:
                    request_mapping_template = file.read()

            response_template_path = build_path(
                artifacts_path,
                resolver_meta.get('response_mapping_template_path'))
            if not artifacts_path or not os.path.exists(response_template_path):
                raise ArtifactError(
                    f"Resolver response mapping template file for type "
                    f"'{type_name}' and field '{field_name}' not found.")
            else:
                with open(response_template_path, 'r',
                          encoding='utf-8') as file:
                    response_mapping_template = file.read()

            resolver_params['request_mapping_template'] = \
                request_mapping_template
            resolver_params['response_mapping_template'] = \
                response_mapping_template

        if max_batch_size := resolver_meta.get('max_batch_size'):
            resolver_params['max_batch_size'] = max_batch_size

        return resolver_params

    @staticmethod
    def _build_function_params_from_meta(func_meta, artifacts_path):
        func_name = func_meta['name']
        _LOG.debug(f"Building parameters for the function '{func_name}'")
        try:
            validate_params(func_name, func_meta, FUNCTION_REQUIRED_PARAMS)
        except ParameterError as e:
            _LOG.warning(str(e))
            _LOG.warning(f"Skipping function '{func_name}'...")
            return

        function_params = {
            'name': func_name,
            'data_source_name': func_meta['data_source_name']
        }

        if description := func_meta.get('description'):
            function_params['description'] = description

        if func_meta.get('runtime') in ('JS', 'APPSYNC_JS'):
            runtime = {
                'name': 'APPSYNC_JS',
                'runtimeVersion': '1.0.0'
            }
            function_params['runtime'] = runtime
        else:
            runtime = None

        if runtime:
            code_path = build_path(artifacts_path,
                                   func_meta.get('code_path'))
            if not artifacts_path or not os.path.exists(code_path):
                raise ArtifactError(
                    f"Function '{func_name}' code file not found.")

            with open(code_path, 'r', encoding='utf-8') as file:
                code = file.read()
            function_params['code'] = code
        else:
            _LOG.debug('Runtime is not JS')
            function_params['function_version'] = \
                func_meta.get('function_version', DEFAULT_FUNC_VTL_VERSION)
            request_template_path = build_path(
                artifacts_path,
                func_meta.get('request_mapping_template_path'))
            if not artifacts_path or not os.path.exists(request_template_path):
                raise ArtifactError(
                    f"Function '{func_name}' request mapping template file "
                    f"not found.")
            else:
                with open(request_template_path, 'r',
                          encoding='utf-8') as file:
                    request_mapping_template = file.read()

            response_template_path = build_path(
                artifacts_path,
                func_meta.get('response_mapping_template_path'))
            if not artifacts_path or not os.path.exists(response_template_path):
                raise ArtifactError(
                    f"Function '{func_name}' response mapping template file "
                    f"not found.")
            else:
                with open(response_template_path, 'r',
                          encoding='utf-8') as file:
                    response_mapping_template = file.read()

            function_params['request_mapping_template'] = \
                request_mapping_template
            function_params['response_mapping_template'] = \
                response_mapping_template

        if max_batch_size := func_meta.get('max_batch_size'):
            function_params['max_batch_size'] = max_batch_size

        return function_params

    def _resolve_log_config(self, log_config: dict, api_name: str) -> dict:
        if log_config.pop('logging_enabled', False):
            _LOG.debug(
                f"Building log_config parameters for the AppSync '{api_name}'"
            )
            log_role_name = log_config.pop('cloud_watch_logs_role_name', None)
            if log_role_name:
                log_config['cloud_watch_logs_role_arn'] = \
                    self._build_iam_role_arn(log_role_name)
            else:
                _LOG.warning(
                    f"Cloud watch logs can't be configured for the AppSync "
                    f"'{api_name}' because 'cloud_watch_logs_role_name' not "
                    f"specified"
                )
                log_config = {}
        else:
            log_config = {}
        return log_config

    def build_graphql_api_arn(self, api_id: str) -> str:
        return f'arn:aws:appsync:{self.appsync_conn.client.meta.region_name}' \
               f':{self.account_id}:apis/{api_id}'

    def describe_graphql_api(self, name, meta, api_id=None):
        if not api_id:
            api = self.appsync_conn.get_graphql_api_by_name(name)
            if not api:
                return {}
            api_id = api['apiId']

        response = {'apiId': api_id}

        arn = self.build_graphql_api_arn(api_id)
        return {
            arn: build_description_obj(response, name, meta)
        }

    def remove_graphql_api(self, args):
        return self.create_pool(self._remove_graphql_api, args)

    @unpack_kwargs
    def _remove_graphql_api(self, arn, config):
        api_id = config['description']['apiId']
        try:
            self.appsync_conn.delete_graphql_api(api_id)
            group_names = self.cw_logs_conn.get_log_group_names()
            for each in group_names:
                if api_id == each.split('/')[-1]:
                    self.cw_logs_conn.delete_log_group_name(each)
            _LOG.info(f'GraphQL API {api_id} was removed.')
            return {arn: config}
        except ClientError as e:
            if e.response['Error']['Code'] == 'NotFoundException':
                _LOG.warning(f'GraphQL API {api_id} is not found')
                return {arn: config}
            else:
                raise e

    def update_graphql_api(self, args):
        return self.create_pool(self._update_graphql_api, args, 1)

    @unpack_kwargs
    def _update_graphql_api(self, name, meta, context):
        extract_to = None
        api = self.appsync_conn.get_graphql_api_by_name(name)
        if not api:
            raise ResourceNotFoundError(f'{name} GraphQL API does not exist.')

        archive_path = meta.get('deployment_package')
        if archive_path:
            extract_to = self._extract_zip(archive_path, name)
        else:
            _LOG.warning('Cannot find appsync deployment package!')
        api_id = api['apiId']
        auth_type = meta.get('primary_auth_type')
        extra_auth_types = meta.get('extra_auth_types', [])
        lambda_auth_config = meta.get('lambda_authorizer_config', {})
        user_pool_config = meta.get('user_pool_config', {})

        if auth_type == AWS_LAMBDA_TYPE:
            lambda_region = lambda_auth_config.pop(AWS_REGION_PARAMETER, None)
            lambda_name = lambda_auth_config.pop('resource_name', None)
            if not lambda_region or not lambda_name:
                lambda_auth_config = {}
                USER_LOG.error(
                    f"Authorization type '{auth_type}' can't be configured "
                    f"for AppSync '{name}' because lambda resource name or "
                    f"aws region is not specified")
            else:
                lambda_arn = self.build_lambda_arn(
                    region=lambda_region,
                    name=lambda_name)
                lambda_auth_config['authorizer_uri'] = lambda_arn

        if auth_type == AWS_CUP_TYPE:
            cup_name = user_pool_config.pop('resource_name', None)
            cup_id = self.cup_conn.if_pool_exists_by_name(cup_name)
            if cup_id:
                user_pool_config['user_pool_id'] = cup_id
            else:
                user_pool_config = {}
                USER_LOG.error(
                    f"Authorization type '{auth_type}' can't be configured "
                    f"for AppSync '{name}' because Cognito User Pool "
                    f"{cup_name} not found")

        updated_extra_auth_types, is_extra_auth_api_key = \
            self._process_extra_auth(extra_auth_types, name)

        self.appsync_conn.update_graphql_api(
            api_id, name, auth_type=auth_type,
            user_pool_config=user_pool_config,
            open_id_config=meta.get('open_id_config'),
            lambda_auth_config=lambda_auth_config,
            log_config=self._resolve_log_config(
                meta.get('log_config', {}), name
            ),
            xray_enabled=meta.get('xray_enabled'),
            extra_auth_types=updated_extra_auth_types)

        if auth_type == 'API_KEY' or is_extra_auth_api_key:
            if not self.get_active_api_keys(api_id):
                api_key_expiration = meta.get('api_key_expiration_days', 7)
                now = time.time()
                api_key_expires = \
                    int(now + api_key_expiration * ONE_DAY_IN_SECONDS)
                self.appsync_conn.create_api_key(
                    api_id, expires=api_key_expires)

        if schema_path := meta.get('schema_path'):
            schema_full_path = build_path(extract_to, schema_path)
            if not extract_to or not os.path.exists(schema_full_path):
                raise ArtifactError(
                    f'\'{schema_full_path}\' file not found for '
                    f'AppSync \'{name}\'')

            with open(schema_full_path, 'r', encoding='utf-8') as file:
                schema_definition = file.read()

            status, details = \
                self.appsync_conn.create_schema(api_id, schema_definition)
            if status != 'SUCCESS':
                error_message = (
                    f"An error occurred when updating schema. "
                    f"Operation status: '{status}'. ")
                if details:
                    error_message += f"Details: '{details}'"
                raise ResourceProcessingError(error_message)
            else:
                _LOG.info(
                    f"Schema of the AppSync '{name}' updated successfully")

        existent_sources = self.appsync_conn.list_data_sources(api_id)
        data_sources_meta = meta.get('data_sources', [])
        for source_meta in data_sources_meta:
            to_create = True
            for source in existent_sources:
                if source['name'] == source_meta['name']:
                    # update an existent one
                    to_create = False
                    params = self._build_data_source_params_from_meta(
                        source_meta)
                    if not params:
                        break
                    _LOG.debug(f"Updating data source '{source['name']}'")
                    self.appsync_conn.update_data_source(api_id, **params)
                    existent_sources.remove(source)
                    break

            # create a new one
            if to_create:
                params = self._build_data_source_params_from_meta(source_meta)
                if not params:
                    continue
                _LOG.debug(f"Creating data source '{source_meta['name']}'")
                self.appsync_conn.create_data_source(api_id, **params)

        for source in existent_sources:
            self.appsync_conn.delete_data_source(api_id, source['name'])

        actual_funcs = []
        existent_funcs = self.appsync_conn.list_functions(api_id)
        funcs_meta = meta.get('functions', [])
        for func_meta in funcs_meta:
            to_create = True
            for func_conf in existent_funcs:
                if func_meta['name'] == func_conf['name']:
                    # update an existent one
                    to_create = False
                    params = self._build_function_params_from_meta(
                        func_meta, extract_to
                    )
                    if not params:
                        break
                    _LOG.debug(f"Updating function '{func_conf['name']}'")
                    actual_funcs.append(
                        self.appsync_conn.update_function(
                            api_id, func_conf['functionId'], params))
                    existent_funcs.remove(func_conf)
                    break
            # create a new one
            if to_create:
                params = self._build_function_params_from_meta(
                    func_meta, extract_to
                )
                if not params:
                    break
                _LOG.debug(f"Creating function '{func_meta['name']}'")
                actual_funcs.append(
                    self.appsync_conn.create_function(api_id, params))

        types = self.appsync_conn.list_types(api_id)
        existent_resolvers = []
        for t in types:
            existent_resolvers.extend(
                self.appsync_conn.list_resolvers(api_id, t['name']))
        resolvers_meta = meta.get('resolvers', [])

        for resolver_meta in resolvers_meta:
            to_create = True
            for resolver in existent_resolvers:
                if resolver['typeName'] == resolver_meta['type_name'] and \
                        resolver['fieldName'] == resolver_meta['field_name']:
                    # update an existent one
                    to_create = False
                    params = self._build_resolver_params_from_meta(
                            resolver_meta, extract_to, actual_funcs)
                    if not params:
                        break
                    _LOG.debug(
                        f"Updating resolver for type '{resolver['typeName']}' "
                        f"and field '{resolver['fieldName']}'")
                    self.appsync_conn.update_resolver(api_id, **params)
                    existent_resolvers.remove(resolver)
                    break

            # create a new one
            if to_create:
                params = self._build_resolver_params_from_meta(
                            resolver_meta, extract_to, actual_funcs)
                if not params:
                    continue
                _LOG.debug(
                    f"Creating resolver for type '{resolver_meta['typeName']}' "
                    f"and field '{resolver_meta['fieldName']}'")
                self.appsync_conn.create_resolver(api_id, **params)

        for resolver in existent_resolvers:
            self.appsync_conn.delete_resolver(api_id,
                                              type_name=resolver['typeName'],
                                              field_name=resolver['fieldName'])

        for func_conf in existent_funcs:
            self.appsync_conn.delete_function(api_id, func_conf['functionId'])

        _LOG.info(f'Updated AppSync GraphQL API {api_id}')
        return self.describe_graphql_api(name=name, meta=meta, api_id=api_id)

    def _extract_zip(self, path: str, name: str):
        from syndicate.core import PROJECT_STATE

        extract_to = PurePath(self.conf_path, ARTIFACTS_FOLDER,
                              name).as_posix()
        artifact_src_path = posixpath.join(
            self.deploy_target_bucket_key_compound,
            PROJECT_STATE.current_bundle, path)

        if not self.s3_conn.is_file_exists(self.deploy_target_bucket,
                                           artifact_src_path):
            raise ArtifactError(
                f"Deployment package for Appsync '{name}' not found by the "
                f"path '{artifact_src_path}'")

        _LOG.info(f'Downloading an artifact for Appsync \'{name}\'')
        with io.BytesIO() as artifact:
            self.s3_conn.download_to_file(
                    bucket_name=self.deploy_target_bucket,
                    key=artifact_src_path,
                    file=artifact)
            with ZipFile(artifact, 'r') as zf:
                zf.extractall(extract_to)
        return extract_to

    def _process_extra_auth(self, extra_auth_types, api_name):
        is_extra_auth_api_key = False
        result = []
        for auth in extra_auth_types:
            if auth['authentication_type'] == AWS_LAMBDA_TYPE:
                lambda_region = \
                    auth.get('lambda_authorizer_config', {}).pop(
                        AWS_REGION_PARAMETER, None)
                lambda_name = \
                    auth.get('lambda_authorizer_config', {}).pop(
                        'resource_name', None)
                if not lambda_region or not lambda_name:
                    USER_LOG.error(
                        f"Authorization type '{AWS_LAMBDA_TYPE}' can't be "
                        f"configured for AppSync '{api_name}' because lambda "
                        f"resource name or aws region is not specified")
                else:
                    lambda_arn = self.build_lambda_arn(
                        region=lambda_region,
                        name=lambda_name)
                    auth['lambda_authorizer_config']['authorizer_uri'] = \
                        lambda_arn
                    result.append(dict_keys_to_camel_case(auth))
            elif auth['authentication_type'] == AWS_CUP_TYPE:
                cup_name = \
                    auth.get('user_pool_config', {}).pop(
                        'resource_name', None)
                cup_id = self.cup_conn.if_pool_exists_by_name(cup_name)
                if not cup_id:
                    USER_LOG.error(
                        f"Authorization type '{AWS_CUP_TYPE}' can't be "
                        f"configured for AppSync '{api_name}' because Cognito "
                        f"User Pool '{cup_name}' not found")
                else:
                    auth['user_pool_config']['user_pool_id'] = cup_id
                    result.append(dict_keys_to_camel_case(auth))
            elif auth['authentication_type'] == 'API_KEY':
                result.append(dict_keys_to_camel_case(auth))
                is_extra_auth_api_key = True
            else:
                result.append(dict_keys_to_camel_case(auth))

        return result, is_extra_auth_api_key

    def get_active_api_keys(self, api_id):
        api_keys = self.appsync_conn.list_api_keys(api_id)
        now = time.time()
        return [api_key for api_key in api_keys if api_key['expires'] > now]



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import concurrent
import traceback
from concurrent.futures import ALL_COMPLETED
from concurrent.futures.thread import ThreadPoolExecutor
from botocore.exceptions import ClientError, BotoCoreError

from syndicate.commons import deep_get
from syndicate.exceptions import SyndicateBaseError
from syndicate.commons.log_helper import get_logger

_LOG = get_logger(__name__)


class BaseResource:

    def create_pool(self, job, parameters, workers=None):
        """ Create resources in pool in sub processes.

        :param workers:
        :type parameters: iterable
        :type job: func
        """
        exceptions = []
        executor = ThreadPoolExecutor(
            workers) if workers else ThreadPoolExecutor()
        futures_dict = {}
        try:
            # futures = [executor.submit(func, i, kwargs) for i in args]
            futures = []
            for param_chunk in parameters:
                param_chunk['self'] = self
                future = executor.submit(job, param_chunk)
                futures.append(future)
                futures_dict[future] = param_chunk
            concurrent.futures.wait(futures, return_when=ALL_COMPLETED)
            responses = {}
            for future in futures:
                try:
                    result = future.result()
                    if result:
                        responses.update(result)
                except Exception as e:
                    param_chunk = futures_dict[future]
                    resource_name = (
                            param_chunk.get('name') or
                            deep_get(
                                param_chunk,
                                ['config', 'resource_name'],
                                'Unknown'))
                    if isinstance(e, (ClientError, BotoCoreError)):
                        exceptions.append(
                            f'When processing the resource {resource_name} {e}'
                        )
                    elif isinstance(e, SyndicateBaseError):
                        exceptions.append(
                            f'When processing the resource {resource_name} '
                            f'occurred {e.__class__.__name__} {e}'
                        )
                    else:
                        exceptions.append(
                            f'When processing the resource {resource_name} '
                            f'occurred an unexpected error '
                            f'({e.__class__.__name__}) {e}'
                        )
                    _LOG.exception(
                        f'An error occurred when processing the resource '
                        f'\'{resource_name}\'. {traceback.format_exc()}'
                    )

            return (responses, exceptions) if exceptions else responses
        finally:
            executor.shutdown(wait=True)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from json import dumps

from boto3 import resource
from botocore.client import Config
from botocore.exceptions import ClientError

from syndicate.commons import deep_get
from syndicate.exceptions import InvalidValueError
from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger(__name__)


@apply_methods_decorator(retry())
class S3Connection(object):
    """ S3 connection class."""

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.region = region
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.resource = resource('s3', self.region,
                                 config=Config(signature_version='s3v4'),
                                 aws_access_key_id=self.aws_access_key_id,
                                 aws_secret_access_key=self.aws_secret_access_key,
                                 aws_session_token=self.aws_session_token)
        self.client = self.resource.meta.client
        _LOG.debug('Opened new S3 connection.')

    def load_file_body(self, bucket_name, key):
        return self.resource.Object(bucket_name, key).get()['Body'].read()

    def download_file(self, bucket_name, key, file_path):
        self.resource.Bucket(bucket_name).download_file(key, file_path)

    def download_to_file(self, bucket_name, key, file):
        self.resource.Bucket(bucket_name).download_fileobj(key, file)

    def upload_file(self, storage, file_name, bucket_name, folder=''):
        """ Upload specific file to s3.

        :type bucket_name: str
        :type file_name: str
        :type storage: str
        :param storage: path (e.g. '/tmp/')
        :param folder: str
        """
        self.client.upload_file(storage + file_name, bucket_name,
                                folder + file_name)

    def upload_single_file(self, path, key, bucket, extra_args=None):
        """ Uploads file just like method above, but allows to specify
        object key as argument

        :param path: path to actual file on os
        :param key: object key, as it will be uploaded to bucket
        :param bucket: just bucket name
        :param extra_args: Extra arguments that may be passed to the client
        operation
        """
        self.client.upload_file(path, bucket, key, ExtraArgs=extra_args)

    def put_object(self, file_obj, key, bucket,
                   content_type, content_encoding=None):
        if content_encoding is None:
            self.resource.Bucket(bucket).put_object(Body=file_obj,
                                                    ContentType=content_type,
                                                    Key=key)
        else:
            self.resource.Bucket(bucket).put_object(
                Body=file_obj,
                ContentType=content_type,
                Key=key,
                ContentEncoding=content_encoding)

    def is_bucket_exists(self, name):
        """ Check if bucket exists by name.

        :type name: str
        """
        res = self.get_list_buckets()
        existing_buckets = [each['Name'] for each in res]
        return name in existing_buckets

    def get_bucket_acl(self, bucket_name):
        try:
            return self.client.get_bucket_acl(Bucket=bucket_name)
        except ClientError as e:
            if 'NoSuchBucket' in str(e):
                pass  # valid exception
            else:
                raise e

    def get_bucket_location(self, bucket_name):
        try:
            return self.client.get_bucket_location(Bucket=bucket_name)
        except ClientError as e:
            if 'NoSuchBucket' in str(e):
                pass  # valid exception
            else:
                raise e

    def get_bucket_policy(self, bucket_name):
        try:
            return self.client.get_bucket_policy(Bucket=bucket_name)
        except ClientError as e:
            if 'NoSuchBucketPolicy' or 'NoSuchBucket' in str(e):
                pass  # valid exception
            else:
                raise e

    def get_bucket_website(self, bucket_name):
        try:
            return self.client.get_bucket_website(Bucket=bucket_name)
        except ClientError as e:
            if 'NoSuchWebsiteConfiguration' or 'NoSuchBucket' in str(e):
                pass  # valid exception
            else:
                raise e

    def is_file_exists(self, bucket_name, key):
        """ Check if file exists.

        :type bucket_name: str
        :type key: str
        """
        return True if self.retrieve_object_metadata(bucket_name, key) else \
            False

    def create_bucket(self, bucket_name, acl=None, location=None):
        """
        :type bucket_name: str
        :param acl: private|public-read|public-read-write|authenticated-read
        :param location: region
        """
        param = dict(Bucket=bucket_name)
        param['ObjectOwnership'] = 'ObjectWriter'
        if acl:
            param['ACL'] = acl
        if not location:
            location = self.region
        valid_location = ['us-west-1', 'us-west-2', 'ca-central-1',
                          'eu-west-1', 'eu-west-2', 'eu-west-3',
                          'eu-central-1',
                          'ap-south-1', 'ap-southeast-1', 'ap-southeast-2',
                          'ap-northeast-1', 'ap-northeast-2', 'sa-east-1',
                          'us-east-2', 'eu-central-1', 'us-east-1',
                          'eu-north-1']
        if location not in valid_location:
            raise InvalidValueError(
                f"Param 'location' has invalid value. Valid locations: "
                f"'{valid_location}'"
            )
        if location != 'us-east-1':  # this is default location
            param['CreateBucketConfiguration'] = {
                'LocationConstraint': location
            }
        self.client.create_bucket(**param)

    def remove_bucket(self, bucket_name, log_not_found_error=True):
        """ Remove bucket by name. To remove bucket it must be empty.
        log_not_found_error parameter is needed for proper log handling in the
        retry decorator
        """
        bucket = self.resource.Bucket(bucket_name)
        bucket_versioning = self.resource.BucketVersioning(bucket_name)
        if bucket_versioning.status == 'Enabled':
            bucket.object_versions.delete()
        else:
            bucket.objects.all().delete()
        bucket.delete()

    def delete_bucket(self, bucket_name):
        self.client.delete_bucket(Bucket=bucket_name)

    def add_lambda_event_source(self, bucket: str, lambda_arn: str,
                                event_source: dict):
        """ Create event notification in the bucket that triggers the lambda
        Note: two identical events can't be configured for two
        separate lambdas in one bucket

        :param bucket:
        :param lambda_arn:
        :param event_source:
            - s3_events: list[str] - list of S3 event types:
                's3:ReducedRedundancyLostObject'
                's3:ObjectCreated:*'
                's3:ObjectCreated:Put'
                's3:ObjectCreated:Post'
                's3:ObjectCreated:Copy'
                's3:ObjectCreated:CompleteMultipartUpload'
                's3:ObjectRemoved:*'
                's3:ObjectRemoved:Delete'
                's3:ObjectRemoved:DeleteMarkerCreated'
            - filter_rules (optional): list[dict] - list of S3 event filters:
                {'Name': 'prefix'|'suffix', 'Value': 'string'}
        """
        config = self.get_bucket_notification(bucket_name=bucket)
        config.pop('ResponseMetadata')

        if 'LambdaFunctionConfigurations' not in config:
            config['LambdaFunctionConfigurations'] = []

        # for some reason filter rule's name value is in uppercase when
        # should be in lower according to the documentation
        for lambda_config in config['LambdaFunctionConfigurations']:
            filter_rules = deep_get(
                lambda_config, ['Filter', 'Key', 'FilterRules'], [])
            for filter_rule in filter_rules:
                filter_rule['Name'] = filter_rule['Name'].lower()

        params = {
            'LambdaFunctionArn': lambda_arn,
            'Events': event_source['s3_events']
        }
        if event_source.get('filter_rules'):
            params.update({
                'Filter': {
                    'Key': {
                        'FilterRules': event_source['filter_rules']
                    }
                }
            })
        # add event notification to remote if it is not already present
        for remote_event_source in config['LambdaFunctionConfigurations']:
            remote_event_source_copy = remote_event_source.copy()
            remote_event_source_copy.pop('Id')
            if remote_event_source_copy == params:
                break
        else:
            config['LambdaFunctionConfigurations'].append(params)

        self.put_bucket_notification(
            bucket_name=bucket, notification_configuration=config)

    def remove_lambda_event_source(self, bucket: str, lambda_arn: str,
                                   event_source: dict):
        """ Remove event notification in the bucket that triggers the lambda
        Note: two identical events can't be configured for two
        separate lambdas in one bucket

        :param bucket:
        :param lambda_arn:
        :param event_source:
            - s3_events: list[str] - list of S3 event types:
                's3:ReducedRedundancyLostObject'
                's3:ObjectCreated:*'
                's3:ObjectCreated:Put'
                's3:ObjectCreated:Post'
                's3:ObjectCreated:Copy'
                's3:ObjectCreated:CompleteMultipartUpload'
                's3:ObjectRemoved:*'
                's3:ObjectRemoved:Delete'
                's3:ObjectRemoved:DeleteMarkerCreated'
            - filter_rules (optional): list[dict] - list of S3 event filters:
                {'Name': 'prefix'|'suffix', 'Value': 'string'}
        """
        config = self.get_bucket_notification(bucket_name=bucket)
        config.pop('ResponseMetadata')

        if 'LambdaFunctionConfigurations' not in config:
            _LOG.info('No lambda event source to remove')
            return

        saved_lambda_configs = []
        for lambda_config in config['LambdaFunctionConfigurations']:
            # for some reason filter rule's name value is in uppercase when
            # should be in lower according to the documentation
            filter_rules = deep_get(
                lambda_config, ['Filter', 'Key', 'FilterRules'], [])
            for filter_rule in filter_rules:
                filter_rule['Name'] = filter_rule['Name'].lower()

            current_lambda = (lambda_config['LambdaFunctionArn'] == lambda_arn)
            same_filters = filter_rules == event_source.get('filter_rules', [])
            same_events = lambda_config['Events'] == event_source['s3_events']
            # save config if something is different from current meta
            if not current_lambda or not same_filters or not same_events:
                saved_lambda_configs.append(lambda_config)

        config['LambdaFunctionConfigurations'] = saved_lambda_configs
        self.put_bucket_notification(
            bucket_name=bucket, notification_configuration=config)

    def get_list_buckets(self):
        response = self.client.list_buckets()
        return response.get('Buckets')

    def add_bucket_policy(self, bucket_name, policy_document):
        """ Attach inline policy to existing bucket.

        :type bucket_name: str
        :type policy_document: str (json)
        """
        if isinstance(policy_document, dict):
            policy_document = dumps(policy_document)
        self.client.put_bucket_policy(
            Bucket=bucket_name,
            Policy=policy_document
        )

    def add_bucket_rule(self, bucket_name, rule_document):
        """
        Creates rule to existing bucket.

        :param bucket_name:
        :param rule_document:
        :return:
        """
        self.client.put_bucket_lifecycle_configuration(
            Bucket=bucket_name,
            LifecycleConfiguration=rule_document
        )

    def enable_website_hosting(self, bucket, index_doc, error_doc):
        """
        :type bucket: str
        :type index_doc: str
        :type error_doc: str
        """
        self.client.put_bucket_website(
            Bucket=bucket,
            WebsiteConfiguration={
                'ErrorDocument': {
                    'Key': error_doc
                },
                'IndexDocument': {
                    'Suffix': index_doc
                }
            }
        )

    def remove_object(self, bucket_name, file_name, mfa=None, version_id=None,
                      request_payer=None):
        """
        :type bucket_name: str
        :type file_name: str
        :type mfa: str
        :type version_id: str
        :type request_payer: str
        :return: response (dict)
        """
        params = dict(Bucket=bucket_name, Key=file_name)
        if mfa:
            params['MFA'] = mfa
        if version_id:
            params['VersionId'] = version_id
        if request_payer:
            params['RequestPayer'] = request_payer
        return self.client.delete_object(**params)

    def list_objects(self, bucket_name, delimiter=None, encoding_type=None,
                     prefix=None, request_payer=None):
        params = dict(Bucket=bucket_name)
        if delimiter:
            params['Delimiter'] = delimiter
        if encoding_type:
            params['EncodingType'] = encoding_type
        if prefix:
            params['Prefix'] = prefix
        if request_payer:
            params['RequestPayer'] = request_payer

        bucket_objects = []
        response = self.client.list_objects(**params)
        marker = response.get('Marker')
        if response.get('Contents'):
            bucket_objects.extend(response.get('Contents'))
        while marker:
            params['Marker'] = marker
            response = self.client.list_objects(**params)
            marker = response.get('Marker')
            if response.get('Contents'):
                bucket_objects.extend(response.get('Contents'))
        return bucket_objects

    def get_bucket_notification(self, bucket_name):
        try:
            return self.client.get_bucket_notification_configuration(
                Bucket=bucket_name
            )
        except ClientError as e:
            if 'AccessDenied' in str(e):
                _LOG.warning(f'{e}. Bucket name - \'{bucket_name}\'.')
            else:
                raise e

    def put_bucket_notification(self, bucket_name, notification_configuration):
        self.client.put_bucket_notification_configuration(
            Bucket=bucket_name,
            NotificationConfiguration=notification_configuration
        )

    def remove_bucket_notification(self, bucket_name):
        self.client.put_bucket_notification_configuration(
            Bucket=bucket_name,
            NotificationConfiguration={}
        )

    def list_keys(self, bucket_name):
        bucket = self.resource.Bucket(bucket_name)
        result = [obj.key for obj in bucket.objects.all()]
        return result

    def get_keys_by_prefix(self, bucket_name, prefix):
        bucket = self.resource.Bucket(bucket_name)
        return [obj.key for obj in bucket.objects.filter(Prefix=prefix)]

    def list_object_versions(self, bucket_name, delimeter=None,
                             encoding_type=None, prefix=None):
        params = dict(Bucket=bucket_name)
        if delimeter:
            params['Delimiter'] = delimeter
        if encoding_type:
            params['EncodingType'] = encoding_type
        if prefix:
            params['Prefix'] = prefix

        bucket_objects = []
        response = self.client.list_object_versions(**params)
        versions = response.get('Versions', [])
        bucket_objects.extend(
            [{
                'Key': i['Key'],
                'VersionId': i['VersionId']
            } for i in versions])
        key_marker = response.get('NextKeyMarker')
        version_marker = response.get('NextVersionIdMarker')
        while key_marker or version_marker:
            if key_marker:
                params['KeyMarker'] = key_marker
            if version_marker:
                params['VersionIdMarker'] = version_marker
            response = self.client.list_object_versions(**params)
            versions = response.get('Versions', [])
            bucket_objects.extend(
                [{
                    'Key': i['Key'],
                    'VersionId': i['VersionId']
                } for i in versions])
        return bucket_objects

    def list_object_markers(self, bucket_name, delimeter=None,
                            encoding_type=None, prefix=None):
        params = dict(Bucket=bucket_name)
        if delimeter:
            params['Delimiter'] = delimeter
        if encoding_type:
            params['EncodingType'] = encoding_type
        if prefix:
            params['Prefix'] = prefix

        bucket_objects = []
        response = self.client.list_object_versions(**params)
        delete_markers = response.get('DeleteMarkers', [])
        bucket_objects.extend(
            [{
                'Key': i['Key'],
                'VersionId': i['VersionId']
            } for i in delete_markers])
        key_marker = response.get('NextKeyMarker')
        version_marker = response.get('NextVersionIdMarker')
        while key_marker or version_marker:
            if key_marker:
                params['KeyMarker'] = key_marker
            if version_marker:
                params['VersionIdMarker'] = version_marker
            response = self.client.list_object_versions(**params)
            delete_markers = response.get('DeleteMarkers', [])
            bucket_objects.extend(
                [{
                    'Key': i['Key'],
                    'VersionId': i['VersionId']
                } for i in delete_markers])
        return bucket_objects

    def retrieve_object_metadata(self, bucket_name, key):
        try:
            return self.client.head_object(Bucket=bucket_name, Key=key)
        except ClientError as e:
            if 'HeadObject' in str(e):
                pass  # valid exception
            else:
                raise e
            return

    def delete_objects(self, bucket_name, objects, mfa=None,
                       request_payer=None):
        params = dict(Bucket=bucket_name, Delete={'Objects': objects})
        if mfa:
            params['MFA'] = mfa
        if request_payer:
            params['RequestPayer'] = request_payer
        return self.client.delete_objects(**params)

    def put_cors(self, bucket_name, rules):
        """
        Puts buckets configuration for existing bucket.
        :param bucket_name: name of the bucket.
        :param rules: list of rules. Each rule may have
            the following attributes: AllowedHeaders, AllowedMethods,
            AllowedOrigins, ExposeHeaders, MaxAgeSeconds;
        :return: None as boto3 does.
        """

        boto_rules = []
        for rule in rules:
            # converting rule to boto format
            for key in rule.keys():
                if isinstance(rule[key], list) \
                        or isinstance(rule[key], int):
                    pass  # expected
                elif isinstance(rule[key], str):
                    rule[key] = [rule[key]]
                else:
                    raise InvalidValueError(
                        f"Value of CORS rule attribute '{key}' has invalid "
                        f"value: '{rule[key]}'. Should be str, int or list"
                    )
            boto_rules.append(rule)

        # boto3 returns None here
        self.client.put_bucket_cors(
            Bucket=bucket_name,
            CORSConfiguration={
                "CORSRules": boto_rules
            }
        )
        _LOG.info('CORS configuration has been set to bucket {0}'.format(
            bucket_name))

    def is_versioning_enabled(self, bucket_name):
        return self.client.get_bucket_versioning(
            Bucket=bucket_name) == 'Enabled'

    def put_public_access_block(self, bucket_name: str,
                                block_public_acls: bool = True,
                                ignore_public_acls: bool = True,
                                block_public_policy: bool = True,
                                restrict_public_buckets: bool = True):
        public_access_block_configuration = {
            'BlockPublicAcls': block_public_acls,
            'IgnorePublicAcls': ignore_public_acls,
            'BlockPublicPolicy': block_public_policy,
            'RestrictPublicBuckets': restrict_public_buckets
        }
        _LOG.info(f'Setting public access block for bucket: {bucket_name} '
                  f'with params: {public_access_block_configuration}')
        self.client.put_public_access_block(
            Bucket=bucket_name,
            PublicAccessBlockConfiguration=public_access_block_configuration
        )
        _LOG.info(f'Public access block was set')

    def put_bucket_acl(self, bucket_name: str, acl: str):
        self.client.put_bucket_acl(
            Bucket=bucket_name,
            ACL=acl
        )
        _LOG.info(f'ACL was set as \'{acl}\'')



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from boto3 import client
from botocore.exceptions import ClientError

from syndicate.exceptions import InvalidValueError
from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger(__name__)


@apply_methods_decorator(retry())
class DocumentDBConnection(object):
    """ DynamoDB class."""

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.region = region
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.client = client('docdb', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new DocumentDB connection.')

    def create_db_cluster(self, identifier, vpc_security_group_ids=None,
                          port=None, availability_zones=None,
                          master_username=None, master_password=None,
                          tags=None):
        """
        Creates a new Amazon DocumentDB cluster.
        """
        params = {
            'DBClusterIdentifier': identifier,
            'Engine': 'docdb'
        }
        if vpc_security_group_ids:
            params['VpcSecurityGroupIds'] = list(vpc_security_group_ids)
        if port:
            params['Port'] = port
        if availability_zones:
            params['AvailabilityZones'] = list(availability_zones)
        if master_username:
            params['MasterUsername'] = master_username
        if master_password:
            params['MasterUserPassword'] = master_password
        if tags:
            params['Tags'] = tags

        response = self.client.create_db_cluster(**params)
        return response['DBCluster'].get('DBClusterIdentifier')

    def create_db_instance(self, instance_identifier, cluster_identifier,
                           instance_class, availability_zone=None, tags=None):
        """
        Creates a new instance.
        """
        params = {
            'DBInstanceIdentifier': instance_identifier,
            'DBInstanceClass': instance_class,
            'DBClusterIdentifier': cluster_identifier,
            'Engine': 'docdb'
        }
        if availability_zone:
            params['AvailabilityZone'] = availability_zone
        if tags:
            params['Tags'] = tags

        response = self.client.create_db_instance(**params)
        return response['DBInstance'].get('DBInstanceIdentifier')

    def delete_db_instance(self, instance_identifier,
                           log_not_found_error=True):
        """
        Deletes a previously provisioned instance.
        log_not_found_error parameter is needed for proper log handling in the
        retry decorator
        """
        response = self.client.delete_db_instance(
            DBInstanceIdentifier=instance_identifier)
        return response['DBInstance'].get('DBInstanceIdentifier')

    def delete_db_cluster(self, cluster_identifier, skip_final_snapshot=True,
                          final_db_snapshot_identifier=None,
                          log_not_found_error=True):
        """
        Deletes a previously provisioned cluster (all automated backups for
        that cluster are deleted and can't be recovered).
        log_not_found_error parameter is needed for proper log handling in the
        retry decorator
        """
        if skip_final_snapshot and final_db_snapshot_identifier:
            raise InvalidValueError(
                'Only one of this parameters must set to \'true\': '
                'skip_final_snapshot, final_db_snapshot_identifier')

        if not skip_final_snapshot and not final_db_snapshot_identifier:
            raise InvalidValueError(
                'One of this parameters must set to \'true\': '
                'skip_final_snapshot, final_db_snapshot_identifier')

        params = {
            'DBClusterIdentifier': cluster_identifier,
            'SkipFinalSnapshot': skip_final_snapshot,
        }
        if final_db_snapshot_identifier:
            params['FinalDBSnapshotIdentifier'] = final_db_snapshot_identifier
        response = self.client.delete_db_cluster(**params)
        return response['DBCluster'].get('DBClusterIdentifier')

    def describe_db_clusters(self, identifier):
        """
        Describes provisioned Amazon DocumentDB clusters.
        """
        params = {
            'DBClusterIdentifier': identifier
        }
        try:
            response = self.client.describe_db_clusters(**params)
        except ClientError as e:
            if e.response['Error']['Code'] == 'DBClusterNotFoundFault':
                _LOG.warn(f'DocumentDB cluster \'{identifier}\' is not found')
                return None
            else:
                raise e
        return response['DBClusters'][0].get('DBClusterIdentifier')

    def describe_db_instances(self, instance_identifier):
        """
        Describes provisioned Amazon DocumentDB instances.
        """
        params = {
            'DBInstanceIdentifier': instance_identifier
        }
        try:
            response = self.client.describe_db_instances(**params)
        except ClientError as e:
            if e.response['Error']['Code'] == 'DBInstanceNotFound':
                _LOG.warn(f'DocumentDB instance \'{instance_identifier}\' is '
                          f'not found')
                return None
            else:
                raise e
        return response['DBInstances'][0].get('DBInstanceIdentifier')




"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import concurrent
import copy
import functools
from concurrent.futures import ALL_COMPLETED, ThreadPoolExecutor
from functools import cmp_to_key
from typing import Any

from syndicate.exceptions import ResourceProcessingError, ProjectStateError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.build.bundle_processor import create_deploy_output, \
    load_deploy_output, load_failed_deploy_output, load_meta_resources, \
    remove_failed_deploy_output, load_latest_deploy_output, \
    remove_deploy_output
from syndicate.core.build.meta_processor import resolve_meta, \
    populate_s3_paths, resolve_resource_name, get_meta_from_output, \
    resolve_tags, preprocess_tags
from syndicate.core.constants import (BUILD_META_FILE_NAME,
                                      CLEAN_RESOURCE_TYPE_PRIORITY,
                                      DEPLOY_RESOURCE_TYPE_PRIORITY,
                                      UPDATE_RESOURCE_TYPE_PRIORITY,
                                      PARTIAL_CLEAN_ACTION, ABORTED_STATUS,
                                      LAMBDA_TYPE, LAMBDA_LAYER_TYPE)
from syndicate.core.helper import prettify_json, strip_prefix_suffix
from syndicate.core.build.helper import assert_bundle_bucket_exists, \
    construct_deploy_s3_key_path

BUILD_META = 'build_meta'
DEPLOYMENT_OUTPUT = 'deployment_output'

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()


def _process_resources(
        resources: list,
        handlers_mapping: dict,
        describe_handlers: dict | None = None,
        pass_context: bool = False,
        output: dict | None = None
) -> tuple[bool, Any]:
    output = output or {}
    errors = []
    args = []
    resource_type = None
    is_succeeded = True
    try:
        for res_name, res_meta in resources:
            current_res_type = res_meta['resource_type']

            if resource_type is None:
                resource_type = current_res_type

            if current_res_type == resource_type:
                args.append(_build_args(name=res_name,
                                        meta=res_meta,
                                        context=output,
                                        pass_context=pass_context))
                continue
            elif current_res_type != resource_type:
                USER_LOG.info(f'Processing {resource_type} resources')
                func = handlers_mapping[resource_type]
                response = func(args)
                response_errors = process_response(response=response,
                                                   output=output)
                errors.extend(response_errors)

                del args[:]
                args.append(_build_args(name=res_name,
                                        meta=res_meta,
                                        context=output,
                                        pass_context=pass_context))
                resource_type = current_res_type
        if args:
            USER_LOG.info(f'Processing {resource_type} resources')
            func = handlers_mapping[resource_type]

            response = func(args)
            response_errors = process_response(response=response,
                                               output=output)
            errors.extend(response_errors)

        if errors:
            is_succeeded = False

    except Exception as e:
        USER_LOG.exception(
            f'Error occurred while {resource_type} resource creating: {e}'
        )
        is_succeeded = False

    if not is_succeeded:
        for res_name, res_meta in resources:
            _LOG.debug(f"Describing the resource '{res_name}'")
            func = describe_handlers[res_meta['resource_type']]
            try:
                response = func(res_name, res_meta)
            except Exception as e:
                _LOG.debug(
                    f"The next error occurred during the resource "
                    f"'{res_name}' describing '{e}'"
                )
                response = ({}, [str(e)])
            if response:
                response_errors = process_response(response=response,
                                                   output=output)
                errors.extend(response_errors)

    return is_succeeded, output


def _process_resources_with_dependencies(resources, handlers_mapping,
                                         describe_handlers, pass_context=False,
                                         overall_resources=None, output=None,
                                         current_resource_type=None,
                                         run_count=0):
    overall_resources = overall_resources or resources
    output = output or {}
    resource_type = None
    is_succeeded = True
    errors = []
    try:
        for res_name, res_meta in resources:
            args = []
            resource_type = res_meta['resource_type']

            if res_meta.get('processed'):
                _LOG.debug(f"Processing of '{resource_type}' '{res_name}' "
                           f"skipped. Resource already processed")
                continue

            if run_count >= len(overall_resources):
                raise ResourceProcessingError(
                    "An infinite loop detected in resource dependencies")

            run_count += 1

            dependencies = [item['resource_name'] for item in
                            res_meta.get('dependencies', [])]
            _LOG.debug(f"'{resource_type}' '{res_name}' depends on resources: "
                       f"{dependencies}")
            # Order of items in depends_on_resources is important!
            depends_on_resources = []
            for dep_res_name in dependencies:
                for overall_res_name, overall_res_meta in overall_resources:
                    if overall_res_name == dep_res_name:
                        depends_on_resources.append((overall_res_name,
                                                     overall_res_meta))

            if depends_on_resources:
                _LOG.info(
                    f"Processing '{resource_type}' '{res_name}' dependencies "
                    f"{prettify_json(res_meta['dependencies'])}")

                success, output = _process_resources_with_dependencies(
                    resources=depends_on_resources,
                    handlers_mapping=handlers_mapping,
                    describe_handlers=describe_handlers,
                    pass_context=pass_context,
                    overall_resources=overall_resources,
                    output=output,
                    current_resource_type=current_resource_type,
                    run_count=run_count)

                if not success:
                    return False, output

            args.append(_build_args(name=res_name,
                                    meta=res_meta,
                                    context=output,
                                    pass_context=pass_context))
            if current_resource_type != resource_type:
                USER_LOG.info(f'Processing {resource_type} resources')
                current_resource_type = resource_type
            func = handlers_mapping[resource_type]
            response = func(args)
            response_errors = process_response(response=response,
                                               output=output)
            errors.extend(response_errors)

            res_meta['processed'] = True
            overall_res_index = overall_resources.index(
                (res_name, res_meta))
            overall_resources[overall_res_index][-1]['processed'] = True

        if errors:
            is_succeeded = False

    except Exception as e:
        if 'An infinite loop' in str(e):
            USER_LOG.error(e.args[0])
        else:
            USER_LOG.exception(f"Error occurred while '{resource_type}' "
                               f"resource creating: {str(e)}")
        is_succeeded = False

    if not is_succeeded:
        for res_name, res_meta in resources:
            _LOG.debug(f"Describing the resource '{res_name}'")
            func = describe_handlers[res_meta['resource_type']]
            try:
                response = func(res_name, res_meta)
            except Exception as e:
                _LOG.debug(
                    f"The next error occurred during the resource "
                    f"'{res_name}' describing '{e}'"
                )
                response = ({}, [str(e)])
            if response:
                response_errors = process_response(response=response,
                                                   output=output)
                errors.extend(response_errors)

    return is_succeeded, output


def _build_args(name, meta, context, pass_context=False):
    """
    Builds parameters to pass to resource_type handler.
    Default parameters dict consists of name and meta keys.
    If pass_context set to True, parameters dict is extended with 'context' key
    :param name: name of the resource
    :param meta: definition of the resource
    :param context: result of previously deployed resources in scope of current syndicate execution
    :param pass_context: flag. Manages if output will be included to parameters
    :return: prepared parameters to be passed to resource_type handler.
    """
    params = {'name': name, 'meta': meta}
    if pass_context:
        params['context'] = context
    return params


def update_failed_output(res_name, res_meta, resource_type, output):
    from syndicate.core import PROCESSOR_FACADE

    try:
        describe_func = PROCESSOR_FACADE.describe_handlers()[resource_type]
        failed_resource_output = describe_func(res_name, res_meta)
        if failed_resource_output:
            if isinstance(failed_resource_output, list):
                for item in failed_resource_output:
                    output.update(item)
            else:
                output.update(failed_resource_output)
    except Exception as e:
        _LOG.warning(f'Unable to describe {resource_type} '
                     f'resource with name {res_name}. Exception: {e}')
    return output


def deploy_resources(
        resources: list,
        output=None,
) -> tuple[bool, Any]:
    from syndicate.core import PROCESSOR_FACADE
    process_with_dependency = False

    for _, res_meta in resources:
        res_priority = DEPLOY_RESOURCE_TYPE_PRIORITY[res_meta['resource_type']]
        dependencies = res_meta.get('dependencies', [])
        dep_priorities = [
            DEPLOY_RESOURCE_TYPE_PRIORITY[item['resource_type']] for item in
            dependencies]

        if dep_priorities:
            if max(dep_priorities) >= res_priority:
                process_with_dependency = True
                break

    if process_with_dependency:
        USER_LOG.warning(
            'Resource dependency with higher deployment priority from a '
            'resource with equal or lower deployment priority detected. '
            'Deployment may take a little bit more time than usual.')

        return _process_resources_with_dependencies(
            resources=resources,
            handlers_mapping=PROCESSOR_FACADE.create_handlers(),
            describe_handlers=PROCESSOR_FACADE.describe_handlers(),
            output=output)

    return _process_resources(
        resources=resources,
        handlers_mapping=PROCESSOR_FACADE.create_handlers(),
        describe_handlers=PROCESSOR_FACADE.describe_handlers(),
        output=output)


def update_resources(
        resources: list[tuple[str, dict]],
        old_resources: set,
) -> tuple[bool, Any]:
    from syndicate.core import PROCESSOR_FACADE
    # exclude new resources that were added after deployment
    to_remove = \
        [i for i, res in enumerate(resources) if res[0] not in old_resources]
    for i in reversed(to_remove):
        USER_LOG.warning(
            f'Skipping resource `{resources[i][0]}` due to absence in initial '
            f'deployment output. If this is an unexpected behaviour, '
            f'please check your configuration.'
        )
        resources.pop(i)

    return _process_resources(
        resources=resources,
        handlers_mapping=PROCESSOR_FACADE.update_handlers(),
        describe_handlers=PROCESSOR_FACADE.describe_handlers(),
        pass_context=True)


def clean_resources(output):
    from syndicate.core import PROCESSOR_FACADE
    clean_output = {}
    errors = []
    args = []
    resource_type = None
    # clean all resources
    for arn, config in output:
        res_type = config['resource_meta']['resource_type']
        if resource_type is None:
            resource_type = res_type

        if res_type == resource_type:
            args.append({'arn': arn, 'config': config})
            continue
        elif res_type != resource_type:
            USER_LOG.info('Removing {0} resources ...'.format(resource_type))
            func = PROCESSOR_FACADE.remove_handlers()[resource_type]
            result = func(args)
            response_errors = process_response(result, clean_output)
            errors.extend(response_errors)
            del args[:]
            args.append({'arn': arn, 'config': config})
            resource_type = res_type
    if args:
        USER_LOG.info('Removing {0} resources ...'.format(resource_type))
        func = PROCESSOR_FACADE.remove_handlers()[resource_type]
        result = func(args)
        response_errors = process_response(result, clean_output)
        errors.extend(response_errors)

    removed_resources_arn = list(clean_output.keys())
    success = False if errors else True

    if not success:
        for arn, config in output:
            res_name = config['resource_name']
            res_meta = config['resource_meta']
            res_type = res_meta['resource_type']

            func = PROCESSOR_FACADE.describe_handlers()[res_type]
            response = func(res_name, res_meta)

            if response and arn in removed_resources_arn:
                removed_resources_arn.remove(arn)
                USER_LOG.warning(
                    f"Resource '{res_name}' of type '{res_type}' was not "
                    f"cleaned."
                )
            elif not response and arn not in removed_resources_arn:
                removed_resources_arn.append(arn)
                USER_LOG.warning(
                    f"Resource '{res_name}' of type '{res_type}' not found; "
                    f"it will be removed from the deployment output."
                )

    return success, removed_resources_arn


def continue_deploy_resources(resources, latest_deploy_output):
    for arn, meta in latest_deploy_output.items():
        for resource_name, resource_meta in resources:
            if resource_name == meta['resource_name']:
                resources.remove((resource_name, resource_meta))

    if not resources:
        USER_LOG.info('Skipping deployment because all specified resources '
                      'already deployed')
        return True, latest_deploy_output, []

    return deploy_resources(resources)


def process_response(
        response: tuple,
        output: dict,
) -> list:
    errors = []

    if isinstance(response, dict):
        output.update(response)
    elif isinstance(response, tuple):
        result, exceptions = response

        if isinstance(result, dict):
            output.update(result)
        else:
            _LOG.warning(
                f"Got unexpected response. Expect dict. Got '{type(response)}',"
                f" '{str(response)}'"
            )

        if isinstance(exceptions, list):
            errors.extend(exceptions)
            for each in exceptions:
                USER_LOG.error(each)
        else:
            USER_LOG.error(str(exceptions))
            errors.append(str(exceptions))

    return errors


def __move_output_content(args, failed_output, updated_output):
    for arg in args:
        resource_output = __find_output_by_resource_name(
            failed_output, arg['name'])
        if resource_output:
            updated_output.update(resource_output)


def __find_output_by_resource_name(output, resource_name):
    found_items = {}
    for k, v in output.items():
        if v['resource_name'] == resource_name:
            found_items[k] = v
    return found_items


def _compare_external_resources(expected_resources):
    from syndicate.core import PROCESSOR_FACADE
    compare_funcs = PROCESSOR_FACADE.compare_meta_handlers()

    errors = {}
    for resource_name, resource_meta in expected_resources.items():
        resource_type = resource_meta.get('resource_type')
        if not (func := compare_funcs.get(resource_type)):
            resource_errors = f'The resource type \'{resource_type}\' is ' \
                              f'not supported for external configuration.'
        else:
            resource_errors = func(resource_name, resource_meta)

        if resource_errors:
            errors[resource_name] = resource_errors

    if errors:
        import os
        error = f'{os.linesep}'.join(errors.values())
        raise ResourceProcessingError(error)


def create_deployment_resources(
        *,
        deploy_name: str,
        bundle_name: str,
        deploy_only_types: tuple | None = None,
        deploy_only_resources: tuple | None = None,
        excluded_resources: tuple | None = None,
        excluded_types: tuple | None = None,
        continue_deploy: bool = False,
        replace_output: bool = False,
        rollback_on_error: bool = False,
) -> bool | str:
    is_ld_output_regular, latest_deploy_output = \
        load_latest_deploy_output(failsafe=True)
    if latest_deploy_output is False:
        USER_LOG.warning(f'The latest deploy output is absent. The current '
                         f'deployment will be performed without taking into '
                         f'account the latest deployment.')
    elif is_ld_output_regular is True:
        _LOG.info(f'The latest deployment has status succeeded. '
                  f'Loaded output:\n {prettify_json(latest_deploy_output)}')
    elif is_ld_output_regular is False:
        _LOG.info(f'The latest deployment has status failed. '
                  f'Loaded output:\n {prettify_json(latest_deploy_output)}')

    resources = load_meta_resources(bundle_name)
    # validate_deployment_packages(resources)
    _LOG.debug(f'{BUILD_META_FILE_NAME} file was loaded successfully')

    resources = resolve_meta(resources)
    _LOG.debug('Names were resolved')
    resources = populate_s3_paths(resources, bundle_name)
    _LOG.debug('Artifacts s3 paths were resolved')
    resolve_tags(resources)

    deploy_only_resources = _resolve_names(deploy_only_resources)
    excluded_resources = _resolve_names(excluded_resources)
    _LOG.info('Prefixes and suffixes of any resource names have been resolved')

    expected_external_resources = {
        key: value for key, value in resources.items() if value.get('external')
    }
    if expected_external_resources:
        _compare_external_resources(expected_external_resources)
        _LOG.info('External resources were matched successfully')

    resources = _filter_resources(
        resources_meta=resources,
        resource_names=deploy_only_resources,
        resource_types=deploy_only_types,
        exclude_names=excluded_resources,
        exclude_types=excluded_types
    )

    if not resources:
        return ABORTED_STATUS

    _LOG.debug(f'Going to create: {resources}')

    # sort resources with priority
    resources_list = list(resources.items())
    resources_list.sort(key=cmp_to_key(compare_deploy_resources))

    _LOG.info('Going to deploy AWS resources')
    if continue_deploy:
        if latest_deploy_output is False:
            USER_LOG.warning(
                f'The latest deploy output is absent. The command will be '
                f'executed without taking into account the '
                f'`--continue_deploy` parameter.')
        success, output = continue_deploy_resources(
            resources_list,
            latest_deploy_output if latest_deploy_output else {})
    else:
        success, output = deploy_resources(resources_list)

    # remove failed output from bucket
    if is_ld_output_regular is False:
        remove_failed_deploy_output(bundle_name, deploy_name)

    if not success:
        tag_success = True
        if rollback_on_error is True:
            USER_LOG.info(
                "Deployment failed, `rollback_on_error` is enabled,"
                " going to clean resources that have been deployed during"
                " deployment process.")
            output_resources_list = list(output.items())
            output_resources_list.sort(
                key=cmp_to_key(_compare_clean_resources))

            if latest_deploy_output:
                deploy_output_names = list(latest_deploy_output.keys())
                rollback_resources_list = \
                    [resource for resource in output_resources_list if
                     resource[0] not in deploy_output_names]
                clean_resources(rollback_resources_list)
            else:
                clean_resources(output_resources_list)

        else:
            _LOG.info('Going to apply post deployment tags')
            tag_success = _apply_post_deployment_tags(output)

            USER_LOG.info('Going to create deploy output')
            output = {**latest_deploy_output, **output} \
                if latest_deploy_output else output
            create_deploy_output(bundle_name=bundle_name,
                                 deploy_name=deploy_name,
                                 output=output,
                                 success=success,
                                 replace_output=replace_output)

    else:
        USER_LOG.info('AWS resources were deployed successfully')

        # apply dynamic changes that uses ARNs
        _LOG.info('Going to apply dynamic changes')
        _apply_dynamic_changes(resources, output)
        USER_LOG.info('Dynamic changes were applied successfully')

        _LOG.info('Going to apply post deployment tags')
        tag_success = _apply_post_deployment_tags(output)

        USER_LOG.info('Going to create deploy output')
        output = {**latest_deploy_output, **output} \
            if latest_deploy_output else output
        create_deploy_output(bundle_name=bundle_name,
                             deploy_name=deploy_name,
                             output=output,
                             success=success,
                             replace_output=replace_output)

    if not (success is False and rollback_on_error is True):
        USER_LOG.info(f'Deploy output for {deploy_name} was created.')
    return success and tag_success


def update_deployment_resources(
        *,
        bundle_name: str,
        deploy_name: str,
        update_only_types: tuple | None = None,
        update_only_resources: tuple | None = None,
        excluded_resources: tuple | None = None,
        excluded_types: tuple | None = None,
        replace_output: bool = False,
        force: bool = False,
) -> bool | str:
    from click import confirm as click_confirm

    is_ld_output_regular, old_output = load_latest_deploy_output(failsafe=True)
    if is_ld_output_regular is None or old_output is False:
        USER_LOG.error('Deployment to update not found.')
        return ABORTED_STATUS
    elif is_ld_output_regular is True:
        _LOG.info(f'The latest deployment has status succeeded. '
                  f'Loaded output:\n {prettify_json(old_output)}')
    elif is_ld_output_regular is False:
        if not force:
            if not click_confirm(
                    "The latest deployment has status failed. "
                    "Do you want to proceed with updating?"):
                return ABORTED_STATUS

            _LOG.warning(
                'Updating resources despite previous deployment failures')
            _LOG.info(f'Loaded output:\n {prettify_json(old_output)}')

    old_resources = get_meta_from_output(old_output)
    old_resources = _resolve_names(tuple(old_resources.keys()))
    resources = load_meta_resources(bundle_name)
    _LOG.debug(prettify_json(resources))

    resources = resolve_meta(resources)
    _LOG.debug('Names were resolved')
    resources = populate_s3_paths(resources, bundle_name)
    _LOG.debug('Artifacts s3 paths were resolved')
    resolve_tags(resources)

    update_only_resources = _resolve_names(update_only_resources)
    _LOG.info(
        'Prefixes and suffixes of any resource names have been resolved.')

    updatable_types = UPDATE_RESOURCE_TYPE_PRIORITY.keys()
    non_updatable_res = set()
    for name in update_only_resources:
        if name in resources and resources[name]['resource_type'] not in updatable_types:
            non_updatable_res.add(name)
    if non_updatable_res:
        non_updatable_res = list(map(strip_prefix_suffix, non_updatable_res))
        USER_LOG.error(
            f'The following resource(s) have a resource type that cannot be '
            f'updated {non_updatable_res}'
        )
        return ABORTED_STATUS

    # Split resources into updatable and non-updatable
    non_updatable_resources = list(
        k for (k, v) in resources.items()
        if v['resource_type'] not in updatable_types
    )
    resources = dict(
        (k, v) for (k, v) in resources.items()
        if v['resource_type'] in updatable_types
    )

    if not (update_only_types or update_only_resources) and non_updatable_resources:
        non_updatable_resources = list(map(strip_prefix_suffix,
                                           non_updatable_resources))
        USER_LOG.warning(
            f'Please note that the following resource(s) will not be updated '
            f'because they have a resource type that cannot be updated '
            f'{non_updatable_resources}')

    resources = _filter_resources(
        resources_meta=resources,
        resource_names=update_only_resources,
        resource_types=update_only_types,
        exclude_names=excluded_resources,
        exclude_types=excluded_types
    )

    if not resources:
        return ABORTED_STATUS

    _LOG.debug(
        f'Going to update the following resources: {prettify_json(resources)}')
    resources_list = list(resources.items())
    resources_list.sort(key=cmp_to_key(_compare_update_resources))

    success, output = update_resources(resources_list, old_resources)

    _LOG.info('Going to updates tags')
    preprocess_tags(output)
    tag_success = _update_tags(old_output, output)

    _LOG.info('Going to resolve updated resources duplication')
    duplicates = _detect_duplicates(old_output, output)
    for duplicate in duplicates:
        old_output.pop(duplicate)

    create_deploy_output(bundle_name=bundle_name,
                         deploy_name=deploy_name,
                         output={**old_output, **output},
                         success=success,
                         replace_output=replace_output)
    if success and tag_success:
        remove_failed_deploy_output(bundle_name, deploy_name)

    return success and tag_success


def remove_deployment_resources(
        deploy_name: str,
        bundle_name: str,
        clean_only_resources: tuple | None = None,
        clean_only_types: tuple | None = None,
        excluded_resources: tuple | None = None,
        excluded_types: tuple | None = None,
        clean_externals: bool = False,
        preserve_state: bool = False,
):
    is_regular_output = True
    externals = {}
    try:
        output = load_deploy_output(bundle_name, deploy_name)
        _LOG.info('Output file was loaded successfully')
    except ProjectStateError:
        try:
            output = load_failed_deploy_output(bundle_name, deploy_name)
            is_regular_output = False
        except ProjectStateError:
            USER_LOG.error('Deployment to clean not found.')
            return ABORTED_STATUS

    new_output = copy.deepcopy(output)

    clean_only_resources = _resolve_names(clean_only_resources)
    excluded_resources = _resolve_names(excluded_resources)
    _LOG.info('Prefixes and suffixes of any resource names have been resolved')

    if not clean_externals:
        new_output = {}
        for k, v in output.items():
            if v['resource_meta'].get('external'):
                externals.update({k: v})
            else:
                new_output.update({k: v})

    new_output = _filter_resources(
        resources_meta=new_output,
        resources_meta_type=DEPLOYMENT_OUTPUT,
        resource_names=clean_only_resources,
        resource_types=clean_only_types,
        exclude_names=excluded_resources,
        exclude_types=excluded_types
    )

    if not new_output:
        return ABORTED_STATUS

    # sort resources with priority
    resources_list = list(new_output.items())
    resources_list.sort(key=cmp_to_key(_compare_clean_resources))
    _LOG.debug(f'Resources to delete: {prettify_json(resources_list)}')
    if resources_list:
        USER_LOG.info('Going to clean AWS resources')
    else:
        _LOG.info('Clean skipped because resources to clean absent')

    success, removed_resources_arn = clean_resources(resources_list)
    _LOG.debug(f'Removed successfully: \'{removed_resources_arn}\'')

    new_output = {k: v for k, v in new_output.items() if k in
                  removed_resources_arn}

    # remove new_output from bucket
    return _post_remove_output_handling(
        deploy_name=deploy_name,
        bundle_name=bundle_name,
        output=output,
        new_output=new_output,
        is_regular_output=is_regular_output,
        success=success,
        externals=externals,
        preserve_state=preserve_state,
    )


def _post_remove_output_handling(
        deploy_name: str,
        bundle_name: str,
        output: dict,
        new_output: dict,
        is_regular_output: bool,
        success: bool,
        externals: dict = None,
        preserve_state: bool = False
) -> bool | dict:
    not_success_msg = 'All resources specified for this operation were ' \
                      'cleaned despite errors when cleaning the resources.'
    if output == new_output:
        if preserve_state:
            return success

        # remove output from bucket
        remove_failed_deploy_output(bundle_name, deploy_name)
        remove_deploy_output(bundle_name, deploy_name)
        if not success:
            USER_LOG.warning(not_success_msg)
            success = True
    else:
        for key, value in new_output.items():
            output.pop(key, None)
        if externals:
            # if only external resources remain
            if set(output.keys()) == set(externals.keys()):
                if preserve_state:
                    return success

                remove_failed_deploy_output(bundle_name, deploy_name)
                remove_deploy_output(bundle_name, deploy_name)
                if not success:
                    USER_LOG.warning(not_success_msg)
                    success = True
                return success

        create_deploy_output(bundle_name=bundle_name,
                             deploy_name=deploy_name,
                             output=output,
                             success=is_regular_output,
                             replace_output=True)

        if not success:
            return success

        return {'operation': PARTIAL_CLEAN_ACTION}
    return success


def _apply_dynamic_changes(resources, output):
    from syndicate.core import PROCESSOR_FACADE
    pool = ThreadPoolExecutor(max_workers=5)
    futures = []
    for name, meta in resources.items():
        resource_type = meta['resource_type']
        apply_changes = meta.get('apply_changes')
        if apply_changes:
            for apply_item in apply_changes:
                change_type = apply_item['apply_type']
                dependency_name = apply_item['dependency_name']
                res_config = resources.get(dependency_name)
                if not res_config:
                    _LOG.debug('Dependency resource {0} is not found, '
                               'skipping the apply'.format(dependency_name))
                else:
                    dependency_type = res_config['resource_type']
                    func = PROCESSOR_FACADE.resource_identifier() \
                        .get(resource_type)
                    if func:
                        resource_output = __find_output_by_resource_name(
                            output, name)
                        identifier = func(name, resource_output)
                        apply_func = PROCESSOR_FACADE.mapping_applier() \
                            .get(change_type)
                        if apply_func:
                            alias = '#{' + name + '}'
                            f = pool.submit(apply_func, alias, identifier,
                                            apply_item)
                            futures.append(f)
                        else:
                            _LOG.warn('Dynamic apply is not defined '
                                      'for {0} type'.format(change_type))
                    else:
                        _LOG.warn('Resource identifier is not defined '
                                  'for {0} type'.format(dependency_type))
            _LOG.info('Dynamic changes were applied to {0}'.format(name))
    concurrent.futures.wait(futures, timeout=None, return_when=ALL_COMPLETED)


def _apply_post_deployment_tags(
        output: dict,
) -> bool:
    from syndicate.core import RESOURCES_PROVIDER
    tags_resource: RESOURCES_PROVIDER = RESOURCES_PROVIDER.tags_api()
    success: bool = tags_resource.safe_apply_tags(output)
    return success


def _update_tags(
        old_output: dict,
        new_output: dict,
) -> bool:
    from syndicate.core import RESOURCES_PROVIDER
    tags_resource: RESOURCES_PROVIDER = RESOURCES_PROVIDER.tags_api()
    success: bool = tags_resource.safe_update_tags(old_output, new_output)
    return success


def compare_deploy_resources(first, second):
    first_resource_type = first[-1]['resource_type']
    second_resource_type = second[-1]['resource_type']
    first_res_priority = DEPLOY_RESOURCE_TYPE_PRIORITY[first_resource_type]
    second_res_priority = DEPLOY_RESOURCE_TYPE_PRIORITY[second_resource_type]
    return _compare_res(first_res_priority, second_res_priority)


def _compare_clean_resources(first, second):
    first_resource_type = first[-1]['resource_meta']['resource_type']
    second_resource_type = second[-1]['resource_meta']['resource_type']
    first_res_priority = CLEAN_RESOURCE_TYPE_PRIORITY[first_resource_type]
    second_res_priority = CLEAN_RESOURCE_TYPE_PRIORITY[second_resource_type]
    return _compare_res(first_res_priority, second_res_priority)


def _compare_update_resources(first, second):
    first_resource_type = first[-1]['resource_type']
    second_resource_type = second[-1]['resource_type']
    first_res_priority = UPDATE_RESOURCE_TYPE_PRIORITY[first_resource_type]
    second_res_priority = UPDATE_RESOURCE_TYPE_PRIORITY[second_resource_type]
    return _compare_res(first_res_priority, second_res_priority)


def _compare_res(first_res_priority, second_res_priority):
    if first_res_priority < second_res_priority:
        return -1
    elif first_res_priority > second_res_priority:
        return 1
    else:
        return 0


def _resolve_names(names):
    from syndicate.core import CONFIG
    preset_name_resolution = functools.partial(resolve_resource_name,
                                               prefix=CONFIG.resources_prefix,
                                               suffix=CONFIG.resources_suffix)
    resolve_n_unify_names = lambda collection: set(
        collection + tuple(map(preset_name_resolution, collection)))

    return resolve_n_unify_names(names or tuple())


def _filter_resources(
        resources_meta: dict,
        resources_meta_type: str = BUILD_META,
        resource_names: set | None = None,
        resource_types: tuple | None = None,
        exclude_names: set | None = None,
        exclude_types: tuple | None = None,
) -> dict:
    """
    Returns filtered resources based on provided names and types.

    `resources_meta_type` parameter could be either BUILD_META or
    DEPLOYMENT_OUTPUT.
    """

    filtered = {}
    resource_names = set() if resource_names is None else set(resource_names)
    resource_types = set() if resource_types is None else set(resource_types)
    exclude_names = set() if exclude_names is None else set(exclude_names)
    exclude_types = set() if exclude_types is None else set(exclude_types)

    _LOG.debug(f"Include resources by name: {list(resource_names) or 'All'}")
    _LOG.debug(f"Include resources by type: {list(resource_types) or 'All'}")
    _LOG.debug(f"Exclude resources by name: {list(exclude_names) or 'None'}")
    _LOG.debug(f"Exclude resources by type: {list(exclude_types) or 'None'}")

    if not any([resource_names, resource_types]):
        filtered = resources_meta
    else:
        if resources_meta_type == BUILD_META:
            filtered = {
                k: v for k, v in resources_meta.items()
                if k in resource_names or v['resource_type'] in resource_types
            }
        elif resources_meta_type == DEPLOYMENT_OUTPUT:
            filtered = {
                k: v for k, v in resources_meta.items()
                if v['resource_name'] in resource_names
                or v['resource_meta']['resource_type'] in resource_types
            }

    for k, v, in copy.deepcopy(filtered).items():
        if resources_meta_type == BUILD_META:
            if k in exclude_names or v['resource_type'] in exclude_types:
                filtered.pop(k)
        elif resources_meta_type == DEPLOYMENT_OUTPUT:
            if (v['resource_name'] in exclude_names
                    or v['resource_meta']['resource_type'] in exclude_types):
                filtered.pop(k)
    if resources_meta_type == BUILD_META:
        meta_source = 'build meta'
        filtered_names = set(map(strip_prefix_suffix, filtered.keys()))
        missing_names = (
                set(map(strip_prefix_suffix, resource_names)) - filtered_names
        )
    else:
        meta_source = 'deployment output'
        filtered_names = set(map(strip_prefix_suffix,
                                 [v['resource_name'] for v in filtered.values()]))
        missing_names = set(map(strip_prefix_suffix, resource_names)) - filtered_names

    if missing_names:
        USER_LOG.warning(
            f'The following resource(s) will be skipped due to absence in '
            f'{meta_source}: {list(missing_names)}. If this is an unexpected '
            f'behaviour, please check the command parameters.')

    if filtered:
        USER_LOG.info(f'The following resource(s) will be processed: '
                      f'{list(filtered_names)}')
    else:
        USER_LOG.warning(
            'No resources to process. Please check the command parameters.')

    return filtered


def is_deploy_exist(bundle_name, deploy_name):
    from syndicate.core import CONN, CONFIG
    assert_bundle_bucket_exists()
    key_compound = construct_deploy_s3_key_path(bundle_name, deploy_name)
    failed_key_compound = construct_deploy_s3_key_path(
        bundle_name, deploy_name, is_failed=True)
    bucket = CONFIG.deploy_target_bucket
    return CONN.s3().get_keys_by_prefix(bucket, key_compound) or \
        CONN.s3().get_keys_by_prefix(bucket, failed_key_compound)


def _detect_duplicates(output: dict, new_output: dict) -> list:
    duplicates = []
    for arn, meta in new_output.items():
        res_type = meta['resource_meta']['resource_type']
        res_name = meta['resource_name']
        if res_type in [LAMBDA_TYPE, LAMBDA_LAYER_TYPE]:
            base_arn = arn[:arn.index(res_name) + len(res_name)]
            duplicates.extend(
                [
                    k for k in output.keys()
                    if k != arn and k.startswith(base_arn)
                ]
            )
    if duplicates:
        _LOG.info(
            f'The next duplicated ARNs will be removed from the output file '
            f'{duplicates}')
    return duplicates



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.core.helper import (
    unpack_kwargs, dict_keys_to_capitalized_camel_case)
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import build_description_obj, \
    assert_required_params, assert_possible_values

_LOG = get_logger(__name__)


class CognitoUserPoolResource(BaseResource):

    def __init__(self, cognito_idp_conn, account_id,
                 region) -> None:
        self.connection = cognito_idp_conn
        self.account_id = account_id
        self.region = region

    def create_cognito_user_pool(self, args):
        """ Create Cognito user pool in pool in sub processes.

        :type args: list
        """
        return self.create_pool(self._create_cognito_user_pool_from_meta,
                                args)

    def describe_user_pool(self, name, meta, pool_id=None):
        if not pool_id:
            pool_id = self.connection.if_pool_exists_by_name(name)
        if not pool_id:
            return {}
        response = self.connection.describe_user_pool(pool_id)
        arn = f'arn:aws:cognito-idp:{self.region}:{self.account_id}:' \
              f'userpool/{pool_id}'
        return {
            arn: build_description_obj(response, name, meta)
        }

    def is_user_pool_exists(self, pool_id):
        try:
            return True if self.connection.describe_user_pool(pool_id) else \
                False
        except ClientError as e:
            if e.response['Error']['Code'] == 'ResourceNotFoundException':
                _LOG.warn(f'Cognito user pool {pool_id} is not found!')
            else:
                raise e

    @unpack_kwargs
    def _create_cognito_user_pool_from_meta(self, name, meta):
        """ Create Cognito user pool for authentication.

        :type name: str
        :type meta: dict
        """
        pool_id = self.connection.if_pool_exists_by_name(name)
        if pool_id:
            _LOG.warn(f'%s cognito user pool exists.', name)
            return self.describe_user_pool(name=name, meta=meta,
                                           pool_id=pool_id)

        _LOG.info('Creating user pool %s', name)
        auto_verified_attributes = meta.get('auto_verified_attributes', [])
        if not isinstance(auto_verified_attributes, list):
            _LOG.warn('Incorrect value for auto_verified_attributes: %s, '
                      'it must be a list',
                      auto_verified_attributes)
            auto_verified_attributes = []
        assert_possible_values(auto_verified_attributes,
                                ['email', 'phone_number'])

        sms_configuration = meta.get('sms_configuration', {})
        if not isinstance(sms_configuration, dict):
            _LOG.warn('Incorrect value for auto_verified_attributes: %s, '
                      'it must be a dict',
                      sms_configuration)
            sms_configuration = {}
        if 'phone_number' in auto_verified_attributes:
            _LOG.warn(f"'phone_number' is inside {auto_verified_attributes}. "
                      f"Hence 'sns_caller_arn' must be in {sms_configuration}")
            assert_required_params(['sns_caller_arn'], sms_configuration)
        sms_configuration = dict_keys_to_capitalized_camel_case(
            sms_configuration)

        username_attributes = meta.get('username_attributes', [])
        if not isinstance(username_attributes, list):
            _LOG.warn('Incorrect value for username_attributes: %s, '
                      'it must be a list',
                      username_attributes)
            username_attributes = []
        assert_possible_values(username_attributes, ['email', 'phone_number'])

        policies = meta.get('password_policy')
        if policies:
            policies = self.__validate_policies(policies)

        pool_id = self.connection.create_user_pool(
            pool_name=name, auto_verified_attributes=auto_verified_attributes,
            sms_configuration=sms_configuration,
            username_attributes=username_attributes, policies=policies,
            tags=meta.get('tags'))

        custom_attributes = meta.get('custom_attributes')
        if custom_attributes:
            self.add_custom_attributes(pool_id, custom_attributes)
        client = meta.get('client')
        if client:
            self.connection.create_user_pool_client(
                user_pool_id=pool_id, **client)
        _LOG.info('Created cognito user pool %s', pool_id)
        return self.describe_user_pool(name=name, meta=meta, pool_id=pool_id)

    def remove_cognito_user_pools(self, args):
        return self.create_pool(self._remove_cognito_user_pools, args)

    @unpack_kwargs
    def _remove_cognito_user_pools(self, arn, config):
        pool_id = config['description']['UserPool']['Id']
        try:
            self.connection.remove_user_pool(pool_id,
                                             log_not_found_error=False)
            _LOG.info('Cognito user pool %s was removed', pool_id)
            return {arn: config}
        except ClientError as e:
            if e.response['Error']['Code'] == 'ResourceNotFoundException':
                _LOG.warn('Cognito user pool %s is not found', pool_id)
                return {arn: config}
            else:
                raise e

    def get_user_pool_id(self, name):
        """
        Returns user pools ID by its name

        :param name: user pools name
        :return: user pools ID
        """
        return self.connection.if_pool_exists_by_name(user_pool_name=name)

    def cognito_resource_identifier(self, name, output=None):
        if output:
            # cognito currently is not located in different regions
            # process only first object
            pool_output = list(output.items())[0][1]
            # find id from the output
            return pool_output['description']['UserPool']['Id']
        return self.connection.if_pool_exists_by_name(name)

    def add_custom_attributes(self, user_pool_id, attributes):
        custom_attributes = []
        for attr in attributes:
            attr['attribute_data_type'] = attr.pop('type')
            custom_attributes.append(dict_keys_to_capitalized_camel_case(attr))
        self.connection.add_custom_attributes(user_pool_id, custom_attributes)

    @staticmethod
    def __validate_policies(policies):
        if not policies.get('minimum_length'):
            policies['minimum_length'] = 6
        if not policies.get('require_uppercase'):
            policies['require_uppercase'] = True
        if not policies.get('require_symbols'):
            policies['require_symbols'] = True
        if not policies.get('require_lowercase'):
            policies['require_lowercase'] = True
        if not policies.get('require_numbers'):
            policies['require_numbers'] = True

        return {'PasswordPolicy': dict_keys_to_capitalized_camel_case(
            policies)}





#  Copyright 2021 EPAM Systems, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from syndicate.core.transform.terraform.converter.tf_lambda_layer_converter import \
    LambdaLayerConverter
from syndicate.commons.log_helper import get_logger
from syndicate.core.transform.build_meta_transformer import \
    BuildMetaTransformer
from syndicate.core.transform.terraform.converter.tf_api_gateway_converter import \
    ApiGatewayConverter
from syndicate.core.transform.terraform.converter.tf_batch_compenv_converter import \
    BatchComputeEnvConverter
from syndicate.core.transform.terraform.converter.tf_batch_jobdef_converter import \
    BatchJobDefConverter
from syndicate.core.transform.terraform.converter.tf_batch_jobqueue_converter import \
    BatchJobQueueEnvConverter
from syndicate.core.transform.terraform.converter.tf_cloud_watch_alram_converter import \
    CloudWatchAlarmConverter
from syndicate.core.transform.terraform.converter.tf_cloud_watch_rule_converter import \
    CloudWatchRuleConverter
from syndicate.core.transform.terraform.converter.tf_cognito_converter import \
    CognitoConverter
from syndicate.core.transform.terraform.converter.tf_dynamo_db_converter import \
    DynamoDbConverter
from syndicate.core.transform.terraform.converter.tf_iam_policy_converter import \
    IamPolicyConverter
from syndicate.core.transform.terraform.converter.tf_iam_role_converter import \
    IamRoleConverter
from syndicate.core.transform.terraform.converter.tf_kinesis_stream_converter import \
    KinesisStreamConverter
from syndicate.core.transform.terraform.converter.tf_lambda_converter import \
    LambdaConverter
from syndicate.core.transform.terraform.converter.tf_s3_bucket_converter import \
    S3BucketConverter
from syndicate.core.transform.terraform.converter.tf_sns_app_converter import \
    SNSApplicationConverter
from syndicate.core.transform.terraform.converter.tf_sns_topic_converter import \
    SNSTopicConverter
from syndicate.core.transform.terraform.converter.tf_sqs_converter import \
    SQSQueueConverter
from syndicate.core.transform.terraform.terraform_template import \
    TerraformTemplate

READ_CAPACITY_UNITS = 'ReadCapacityUnits'
WRITE_CAPACITY_UNITS = 'WriteCapacityUnits'

_LOG = get_logger(__name__)


class TerraformTransformer(BuildMetaTransformer):

    def __init__(self, bundle_name=None):
        super().__init__(bundle_name)
        self.template = TerraformTemplate('aws', 'default', self.config.region)

    def add_resource(self, transformed_resource):
        pass  # TODO implement or design other approach

    def output_file_name(self) -> str:
        return 'terraform_template.tf.json'

    def _transform_iam_managed_policy(self, name, resource):
        self.convert_resources(resource=resource,
                               converter_type=IamPolicyConverter,
                               name=name)

    def _transform_iam_role(self, name, resource):
        self.convert_resources(resource=resource,
                               converter_type=IamRoleConverter,
                               name=name)

    def _transform_lambda(self, name, resource):
        self.convert_resources(resource=resource,
                               converter_type=LambdaConverter,
                               name=name)

    def _transform_dynamo_db_table(self, name, resource):
        self.convert_resources(resource=resource,
                               converter_type=DynamoDbConverter,
                               name=name)

    def _transform_s3_bucket(self, name, resource):
        self.convert_resources(resource=resource,
                               converter_type=S3BucketConverter,
                               name=name)

    def _transform_cloud_watch_rule(self, name, resource):
        self.convert_resources(resource=resource,
                               converter_type=CloudWatchRuleConverter,
                               name=name)

    def _transform_api_gateway(self, name, resource):
        self.convert_resources(resource=resource,
                               converter_type=ApiGatewayConverter,
                               name=name)

    def _transform_sns_topic(self, name, resource):
        self.convert_resources(resource=resource,
                               converter_type=SNSTopicConverter,
                               name=name)

    def _transform_sqs_queue(self, name, resource):
        self.convert_resources(resource=resource,
                               converter_type=SQSQueueConverter,
                               name=name)

    def _transform_cloudwatch_alarm(self, name, resource):
        self.convert_resources(resource=resource,
                               converter_type=CloudWatchAlarmConverter,
                               name=name)

    def _transform_kinesis_stream(self, name, resource):
        self.convert_resources(resource=resource,
                               converter_type=KinesisStreamConverter,
                               name=name)

    def _transform_sns_application(self, name, resource):
        self.convert_resources(resource=resource,
                               converter_type=SNSApplicationConverter,
                               name=name)

    def _transform_cognito(self, name, resource):
        self.convert_resources(resource=resource,
                               converter_type=CognitoConverter,
                               name=name)

    def _transform_batch_compenv(self, name, resource):
        self.convert_resources(resource=resource,
                               converter_type=BatchComputeEnvConverter,
                               name=name)

    def _transform_batch_jobqueue(self, name, resource):
        self.convert_resources(resource=resource,
                               converter_type=BatchJobQueueEnvConverter,
                               name=name)

    def _transform_batch_jobdef(self, name, resource):
        self.convert_resources(resource=resource,
                               converter_type=BatchJobDefConverter,
                               name=name)

    def _transform_lambda_layer(self, name, resource):
        self.convert_resources(resource=resource,
                               converter_type=LambdaLayerConverter,
                               name=name)

    def convert_resources(self, name, resource, converter_type):
        converter = converter_type(template=self.template, config=self.config,
                                   resources_provider=self.resources_provider)
        converter.convert(name=name, resource=resource)

    def _compose_template(self):
        return self.template.compose_resources()
